from functools import reduce
from operator import mul
from typing import List, Tuple, Callable, Optional, Iterator

import numpy as np


from coker.dynamics.types import (
    TranscriptionOptions,
    ControlVariable,
    PiecewiseConstantVariable,
    SpikeVariable,
)


def legendre_coefficient(n, k):
    """Compute the coefficient of the kth Legendre polynomial of degree n."""
    if k > n:
        return 0
    if (n, k) == (0, 0):
        return 1
    if (n, k) == (1, 1):
        return 1
    if k <= 1:
        if (k == 1 and n % 2 == 0) or (k == 0 and n % 2 == 1):
            return 0
        coeff = -(n + k - 1) / (n - k)
        return coeff * legendre_coefficient(n - 2, k)

    coeff = -((n - k + 2) * (n + k - 1) / (k * (k - 1)))
    return coeff * legendre_coefficient(n, k - 2)


def expand_coefficients(roots: List[float]):
    """Expand polynomials from root form to coefficient form."""
    # p(x) = (x - x_0) (x - x_1) ...  (x  - x_n)
    #
    # final case:
    #   p(x) = [] , r(x) = [a_0, a_1, ... a_n, 1]
    #   where r(x) = sum_i(x^i a_i)
    #   -> return r(x)

    # (x - r_0)(x - r_1) (\sum c_i x^i)
    #  -> (x - r_0)( \sum( (r_1 c_i + c_{i-1}) x^i)

    # initial case: p(x) = coeffs, r(x) = [1]
    coeffs = [1]
    while roots:
        root = roots.pop()
        coeffs = [root * coeffs[0]] + [
            root * c + c_last for c, c_last in zip(coeffs[1:] + [0], coeffs)
        ]
    return coeffs


def lgr_points(n: int) -> List[float]:
    """Compute LGR collocation points for a given order n."""
    assert n > 0
    # Points are roots of P_n(x) + P_{n-1}(x)
    # Where P_n(x) is the nth Legendre Polynomial

    companion_matrix = np.diag(np.ones(n - 1), 1)
    leading_term = legendre_coefficient(n, n)
    for i in range(n):
        companion_matrix[-1, i] = (
            -(legendre_coefficient(n - 1, i) + legendre_coefficient(n, i))
            / leading_term
        )
    roots = np.linalg.eigvals(companion_matrix).tolist()
    roots.append(1)
    roots.sort()

    return [-1] + roots[1:]


def evaluate_legendre_polynomial(x, n):
    return np.polynomial.legendre.legval(x, [0] * n + [1])


def generate_discritisation_operators(
    interval: Tuple[float, float], n: int
) -> Tuple[
    List[float],
    Callable[[float], float],
    List[np.ndarray],
    List[np.ndarray],
    np.ndarray,
]:
    """
    Generates discretisation operators over an interval using Legendre-Gauss-Radau (LGR)
    collocation points. This function computes knot points, Legendre polynomial bases,
    as well as derivative and integral operators for numerical methods.

    Args:
        interval: A tuple representing the interval [a, b] over which the discretisation
            is computed.
        n: The number of discretisation points.

    Returns:
        Tuple containing:
            List[float]: Knot points of length (n + 1).
            Callable[[float], float]: Transformed time variable.
            List[np.ndarray]: Legendre polynomial bases at each point.
            List[np.ndarray]: Derivative operators at each knot point.
            np.ndarray: Integral operator for the interval.
    """
    colocation_times = np.array(lgr_points(n))

    # using n LRG collocation points covering [-1, 1)
    # plus an additional non-collocated point a +1
    #

    bases = np.empty((n + 1, n + 1))
    time_scaling_factor = (interval[1] - interval[0]) / 2
    colocation_coeff = np.zeros((n + 1, n + 1))
    continuity_coeff = np.zeros(n + 1)
    quad_coeff = np.zeros_like(continuity_coeff)

    for i in range(n + 1):
        tau_i = colocation_times[i]
        factors = [
            np.poly1d([1, -tau_j]) / (tau_i - tau_j)
            for tau_j in colocation_times
            if tau_i != tau_j
        ]
        basis_i = reduce(mul, factors)
        bases[:, i] = basis_i.c[::-1]
        dbasis_i = np.polyder(basis_i)

        continuity_coeff[i] = basis_i(1)
        colocation_coeff[i, :] = [
            dbasis_i(tau_j) / time_scaling_factor for tau_j in colocation_times
        ]
        quad_coeff[i] = np.polyint(basis_i)(1.0) * time_scaling_factor
    # see https://mathworld.wolfram.com/RadauQuadrature.html
    weights = (
        np.array(
            [2 / n**2]
            + [
                (1 - x_i) / (n * evaluate_legendre_polynomial(x_i, n - 1)) ** 2
                for x_i in colocation_times[1:-1]
            ]
            + [0]
        )
        * time_scaling_factor
    )

    t = lambda tau: (interval[0] + interval[1]) / 2 + tau * time_scaling_factor

    derivative: List[np.ndarray] = [row for row in colocation_coeff.T]
    bases: List[np.ndarray] = [row for row in bases]
    integral_weights: np.ndarray = weights.reshape(1, n + 1)
    return colocation_times, t, bases, derivative, integral_weights


def split_at_non_differentiable_points(
    control_variables: List[ControlVariable],
    t_final: float,
    transcription_options: TranscriptionOptions,
    additional_points: Optional[List[float]] = None,
) -> List[Tuple[float, float]]:
    """
    Splits the time domain into intervals taking into account control variables, additional
    integration points, and transcription requirements.

    The function determines interval boundaries based on the characteristics and constraints
    of the provided control variables. It also considers any additional points passed, ensuring
    subdivision complies with the transcription options. The intervals are adjusted to guarantee
    a minimum required number and their lengths are subdivided iteratively when necessary.

    Args:
        control_variables (List[ControlVariable]): A list of control variables, which define
            parameters affecting the differentiation process. This may include variables
            such as piecewise constants or spike events.
        t_final (float): The total duration of the time domain or integration window.
        transcription_options (TranscriptionOptions): Configuration options specifying
            transcription constraints, such as the minimum number of intervals.
        additional_points (Optional[List[float]]): Additional time points that should be
            included as interval boundaries, if provided.

    Returns:
        List[Tuple[float, float]]: A list of tuples representing the sorted interval boundaries.
            Each tuple contains the start and end of an interval.
    """
    interval_boundaries = (
        set(additional_points) if additional_points else set()
    )
    for d in control_variables:
        if isinstance(d, PiecewiseConstantVariable):
            assert d.sample_rate > 0, "Sample rate must be positive"
            steps = t_final * d.sample_rate
            interval_boundaries |= {
                i * t_final / steps for i in range(int(steps))
            }

        if isinstance(d, SpikeVariable):
            assert (
                0 <= d.time < t_final
            ), "Spike time must be within integration window"

            interval_boundaries.add(d.time)

    if 0 not in interval_boundaries:
        interval_boundaries.add(0)

    if t_final not in interval_boundaries:
        interval_boundaries.add(t_final)

    sorted_boundaries = list(interval_boundaries)

    sorted_boundaries.sort()

    while len(sorted_boundaries) < transcription_options.minimum_n_intervals:
        intervals = [
            (stop - start, (stop + start) / 2)
            for start, stop in zip(
                sorted_boundaries[:-1], sorted_boundaries[1:]
            )
        ]
        max_length = max(l for l, _ in intervals)

        if all(length == max_length for length, _ in intervals):
            sorted_boundaries += [mid for _, mid in intervals]
        else:
            sorted_boundaries += [
                point for length, point in intervals if point == max_length
            ]
        sorted_boundaries.sort()

    return [
        (start, stop)
        for start, stop in zip(sorted_boundaries[:-1], sorted_boundaries[1:])
    ]


class InterpolatingPoly:
    """Represents a multidimensional polynomial interpolation over a specified
    interval.

    This class facilitates the computation and evaluation of an interpolating
    polynomial given a set of values, the polynomial degree, and its interval.
    It provides functionality to obtain knot points, start and end points, and
    evaluate the polynomial at a given point in the interval, while maintaining
    information about discrete operators and transformations.

    Legendre-Gauss-Radau collocation points are used as the discritisation
    scheme.

    Attributes:
        interval (Tuple[float, float]): The interval [a, b] over which the
            polynomial is defined.
        dimension (int): Dimensionality of the interpolated values.
        degree (int): Degree of the interpolating polynomial.
        values (np.ndarray): The values to interpolate, corresponding to knot
            points of the polynomial.
        s (np.ndarray): Array of discrete points in polynomial parameter space.
        s_to_interval (Callable[[float], float]): Function to map points from
            parameter space to the original interval.
        integrals (np.ndarray): Integration operator for the interpolating
            polynomial.
        width (float): Half-width of the interval, used for transformations
            between parameter space and interval.
        bases (np.ndarray): Basis functions of the interpolating polynomial.
        derivatives (List[np.ndarray]): Derivative basis functions of the
            polynomial.
    """

    def __init__(self, dimension, interval, degree, values):
        """
        Initializes an instance of the class with given parameters and specific configurations
        required for discretization and calculations.

        Args:
            dimension: The dimension of the problem that defines the size of matrices and reshaping.
            interval: The interval [lower_bound, upper_bound] over which discretization is computed.
            degree: The degree of the polynomial basis used for computation.
            values: A numpy array representing the knot points, stacked in order.

        Raises:
            AssertionError: If the shape of the `values` input does not match the expected shape
                calculated from the interval, dimension, and degree parameters.
        """
        self.interval = interval
        self.dimension = dimension
        self.degree = degree
        self.values = values

        op_values = generate_discritisation_operators(interval, degree)

        self.s, self.s_to_interval, bases, derivatives, self.weights = (
            op_values
        )
        self.width = (self.interval[1] - self.interval[0]) / 2
        size = len(self.s) * dimension
        if len(values.shape) == 1:
            values = values.reshape(size, 1)
        assert values.shape == (
            size,
            1,
        ), f"Expected shape ({len(self.s) * dimension}, 1), got {values.shape}"

        self.bases = np.vstack(
            [np.reshape(np.array(base), (1, len(self.s))) for base in bases]
        )
        self.derivatives = [
            np.reshape(np.array(d[:-1]), (1, len(self.s) - 1))
            for d in derivatives
        ]

    def size(self) -> int:
        return len(self.s) * self.dimension

    def _interval_to_s(self, t):
        mean = (self.interval[1] + self.interval[0]) / 2
        return (t - mean) / self.width

    def knot_times(self) -> List[float]:
        # we skip the end point
        return [self.s_to_interval(s_i) for s_i in self.s]

    def start_point(self) -> Tuple[float, np.ndarray]:
        return self.s_to_interval(self.s[0]), self.values[: self.dimension]

    def end_point(self) -> Tuple[float, np.ndarray]:
        return self.s_to_interval(self.s[-1]), self.values[-self.dimension :]

    def as_raw(self) -> np.ndarray:
        t = np.reshape(np.array(self.knot_times()), (-1, 1))
        x = np.reshape(self.values, (-1, self.dimension))
        return np.hstack([t, x]).T

    def knot_points(self) -> Iterator[Tuple[float, np.ndarray, np.ndarray]]:
        """Get the knot points of the interpolating polynomial, not including the end point."""
        # we skip the end point
        t_i = self.knot_times()[:-1]
        n = len(self.s)
        x_i = [
            self.values[i * self.dimension : (i + 1) * self.dimension]
            for i in range(n - 1)
        ]
        dx_i = []
        ds_dt = 1 / self.width
        for s in self.s:

            ds = ds_dt * np.array(
                [i * s ** (i - 1) if i > 0 else 0 for i in range(len(self.s))]
            ).reshape((1, n))
            projection = (ds @ self.bases).T

            # this treats each column as the interpolation point.
            v = self.values.reshape((self.dimension, -1))
            value = v @ projection
            dx_i.append(value)

        for t, x, dx in zip(t_i, x_i, dx_i):
            yield t, x, dx

    def __call__(self, t) -> np.ndarray:
        """Evaluate the interpolating polynomial at time t."""

        assert (
            self.interval[0] <= t <= self.interval[1]
        ), f"Value {t} is not in interval {self.interval}"

        s = self._interval_to_s(t)
        try:
            i = next(i for i, s_i in enumerate(self.s) if abs(s_i - s) < 1e-9)
            return np.reshape(
                self.values[i * self.dimension : (i + 1) * self.dimension],
                (self.dimension,),
            )
        except StopIteration:
            pass
        n = len(self.s)
        s_vector = np.vstack([s**i for i in range(n)])

        projection = s_vector.T @ self.bases

        value = projection @ np.reshape(self.values, (-1, self.dimension))

        return np.reshape(value, (self.dimension,))

    def map(self, func):
        values = np.concatenate(
            [
                func(
                    self.s_to_interval(s),
                    self.values[i * self.dimension : (i + 1) * self.dimension],
                )
                for i, s in enumerate(self.s)
            ]
        )
        try:
            (size,) = values.shape
        except ValueError as ex:
            if len(values.shape) == 2 and values.shape[1] == 1:
                values = np.squeeze(values)
                (size,) = values.shape
            else:
                ex.add_note(
                    f"Cannot reshape values of shape {values.shape}, expected a vector"
                )
                raise ex

        dimension = size // len(self.s)

        return InterpolatingPoly(dimension, self.interval, self.degree, values)



import dataclasses
import enum
from typing import Optional, List, Tuple, Callable

import numpy as np

from coker import Dimension
from coker.algebra.kernel import Tape, Tracer, VectorSpace, Scalar
from coker.backends import get_backend_by_name, get_current_backend


@dataclasses.dataclass
class SolverOptions:
    warm_start: bool = False


class Minimise:
    def __init__(self, expression: Tracer):
        self.expression = expression


class MathematicalProgram:
    def __init__(
        self,
        input_shape: Tuple[Dimension, ...],
        output_shape: Tuple[Dimension, ...],
        impl: Callable,
    ):

        self.input_shape = input_shape
        self.output_shape = output_shape
        self.impl = impl

    def __call__(self, *args):
        assert len(args) == len(self.input_shape)

        return [
            np.reshape(o, dim.shape)
            for o, dim in zip(self.impl(args), self.output_shape)
        ]


class ProblemBuilder:
    def __init__(self, arguments: Optional[List[VectorSpace | Scalar]] = None):

        self.arguments = (
            [self.tape.input(a) for a in arguments] if arguments else []
        )
        self.objective = None
        self.tape: Optional[Tape] = None
        self.constraints = []
        self.outputs = []
        self.initial_conditions = {}
        self.warm_start = False

    def new_variable(self, name, shape=None, initial_value=None):
        assert self.tape is not None
        if shape is None:
            v = self.tape.input(Scalar(name))
            initial_value = 0 if initial_value is None else initial_value
        else:
            v = self.tape.input(VectorSpace(name, shape))
            initial_value = (
                np.zeros(shape=shape)
                if initial_value is None
                else initial_value
            )

        self.initial_conditions[v.index] = initial_value
        return v

    @property
    def input_shape(self) -> Tuple[Dimension, ...]:
        return tuple(i.dim for i in self.arguments)

    @property
    def output_shape(self) -> Tuple[Dimension, ...]:
        return tuple(o.dim for o in self.outputs)

    def build(self, backend: Optional[str] = None) -> MathematicalProgram:
        assert isinstance(self.objective, Minimise)
        assert self.tape is not None
        assert self.outputs

        backend = (
            get_backend_by_name(backend)
            if backend is not None
            else get_current_backend()
        )

        impl = backend.build_optimisation_problem(
            self.objective.expression,  # cost
            self.constraints,
            self.arguments,
            self.outputs,
            self.initial_conditions,
        )

        return MathematicalProgram(self.input_shape, self.output_shape, impl)

    def __enter__(self):
        self.tape = Tape()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.tape = None
        pass


class VariationalProblem:
    """

    A Variational problem is of the form


        min_[u(t), p]  c(1)

        s.t.
            dx/dt = f(x(t), p)                              [Dynamics]
            dq/dt = g(x(t), p)                              [Integral Constraints]
            0 < h(t, x, p)                                  [Path Constraints]
            c = C(t, x, q, p)                               [Cost]
            (t, x, q, u, p) in  [0, 1] * X * Q * U * P      [Set Constraints]


    A solution is the
        - value c,
        - vector function u(t)
        - parameters p

    that (locally) minimise c.


    This is solved by:
        1. Choosing a discritisation scheme for t, [x,q] and u
        2. transcribing the problem into a nonlinear program by evaluating the constraints at knot points
           and adding additional constraints to enforce continuity.
        3. Classifying the NLP and passing it into an appropriate solver.

    """

    def __init__(self, intervals: List[float]):
        self.intervals = intervals
        self.signals = []
        self.variables = []


def norm(arg, order=2):
    pass



from coker.algebra.tensor import SymbolicVector
from coker.algebra.dimensions import Dimension
from coker.algebra.ops import OP

from coker.algebra.helpers import is_scalar
from coker.algebra.factories import zeros



import pathlib
from typing import Type, Callable, Tuple, Dict
from coker.modelling import Block
from coker.algebra.kernel import Function
from coker.helpers import get_all_subclasses

__ALL__ = [
    "Workspace",
]


class ComponentHandle:
    def __init__(self, path, in_ports, out_ports):
        self.path = path
        self.in_ports = in_ports
        self.out_ports = out_ports

    @property
    def input(self):
        if len(self.in_ports) == 1:
            return self.in_ports[0]

        return self.in_ports

    @property
    def output(self):
        if len(self.out_ports) == 1:
            return self.out_ports[0]

        return self.out_ports


class Worksheet:

    def __init__(self):
        self.components = []
        self.connections = []
        self.input_ports = set()

    def add_instance(self, constructor: Callable, *args, name=None, **kwargs):
        try:
            component: coker.Block = constructor(*args, **kwargs)
        except Exception as e:
            raise e
        if not isinstance(component, Block):
            raise TypeError("Component must be a subclass of `coker.Block`")

        index = len(self.components)
        self.components.append(component)

        component_path = (
            f"/{name}" if name else f"/{component.__class__.__name__}{index}"
        )
        in_ports = [f"{component_path}/in_{i}" for i in component.spec.inputs]
        out_ports = [
            f"{component_path}/in_{i}" for i in component.spec.outputs
        ]
        handle = ComponentHandle(component_path, in_ports, out_ports)

        return handle

    @property
    def free_parameters(self):
        for component in self.components:
            component.free_parameters()


class Simulation(Worksheet):
    def __init__(self, window: Tuple[float | int, float | int], step_size):
        super().__init__()
        self.window = window
        self.probes = []
        self.step_size = step_size

    def add_probe(self, signal, *args, renderer=None, **kwargs):
        self.probes.append(signal)

    def run(self):
        # compile model into dae
        # 0 = Af(t, x, u, p ) - u
        # 0 = Phi(t, dx, x, u, p)
        pass


class BuildRecipe(Worksheet):
    def __init__(self, name_space: str):
        self.name_space: str = name_space
        self.kernels: Dict[str, Function] = {}

    def add_function(self, name: str, function: Function):
        self.kernels[name] = function


def get_worksheets():
    build_recipes = get_all_subclasses(BuildRecipe)
    sim_recipes = get_all_subclasses(Simulation)

    return [("Deploy", name, value) for name, value in build_recipes] + [
        ("Simulation", name, value) for name, value in sim_recipes
    ]



import abc
from typing import List, Callable, Optional, Tuple, Union, Iterator
from dataclasses import dataclass, field

from coker.algebra.kernel import (
    Dimension,
    FunctionSpace,
    Scalar,
    VectorSpace,
    Function,
    Noop,
    InequalityExpression,
    function,
)
import numpy as np
from typing import Dict


@dataclass
class DynamicsSpec:
    inputs: FunctionSpace
    parameters: Scalar | VectorSpace
    algebraic: Optional[VectorSpace]

    initial_conditions: Callable[
        [VectorSpace, VectorSpace], Tuple[np.ndarray, np.ndarray]
    ]
    """ [x, z] = initial_conditions(t_0, p) """

    dynamics: Callable[
        [float, np.ndarray, np.ndarray, np.ndarray, np.ndarray], np.ndarray
    ]
    """dx = dynamics(t, x, z, u, p)"""

    constraints: Callable[
        [float, np.ndarray, np.ndarray, np.ndarray, np.ndarray], np.ndarray
    ]
    """ g(t, x, z, u, p) = constraints(t, x, z, u, p) = 0."""

    outputs: Callable[
        [float, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray],
        np.ndarray,
    ]
    """ y(t) = outputs(t, x, z, u, p, q) """

    quadratures: Callable[
        [float, np.ndarray, np.ndarray, np.ndarray], np.ndarray
    ]
    """ dq/dt = quadratures(t, x, u, p) """


@dataclass
class DynamicalSystem:
    inputs: FunctionSpace
    parameters: VectorSpace | Scalar
    x0: Function
    dxdt: Function
    g: Optional[Function]
    dqdt: Optional[Function]
    y: Function
    solver_parameters: Optional[object] = field(default=None)

    def get_state_dimensions(self) -> Tuple[Dimension, Dimension, Dimension]:
        _t, x_dim, z_dim, _u, _p, q_dim = self.y.input_shape()
        return x_dim, z_dim, q_dim

    def backend(self):
        return self.dxdt.backend

    def _map_arguments(self, *args):
        arg_stack = list(reversed(args))
        t = arg_stack.pop()
        try:
            u = (
                arg_stack.pop()
                if self.inputs is not Noop() or len(args) == 3
                else None
            )
            p = (
                arg_stack.pop()
                if self.parameters is not None or len(args) == 3
                else None
            )
        except IndexError as ex:
            raise ValueError(
                f"Invalid number of arguments: Expected 2 - 3, received: {len(args)}"
            ) from ex

        return t, u, p

    def __call__(self, *args):
        from coker.backends import get_backend_by_name

        t, u, p = self._map_arguments(*args)

        x0, z0 = self.x0(0, u, p)

        # solve ODE
        # x' = dxdt(...)
        # 0  = g(...)
        # to get x,z over the interval

        if self.dqdt is not Noop():
            # zeros, the same size a q
            raise NotImplementedError
        else:
            q0 = None

        backend = get_backend_by_name(self.dxdt.backend)
        x, z, q = backend.evaluate_integrals(
            [self.dxdt, self.g, self.dqdt],
            [x0, z0, q0],
            t,
            [u, p],
            solver_parameters=self.solver_parameters,
        )

        if isinstance(t, (float, int)):
            return self.y(t, x, z, u, p, q)

        def map_args(i):
            x_i = x[:, i]
            z_i = z[:, i] if z is not None else None
            q_i = q[:, i] if q is not None else None
            return x_i, z_i, u, p, q_i

        if self.y.output_shape()[0].is_scalar() or self.y.output_shape()[
            0
        ].dim == (1,):
            y = np.concatenate(
                [self.y(t_i, *map_args(i)) for i, t_i in enumerate(t)]
            )
        else:
            y = np.vstack(
                [self.y(t_i, *map_args(i)) for i, t_i in enumerate(t)]
            )

        return y

    def output_as_function_space(self) -> FunctionSpace:
        t, _x_dim, _z_dim, u, p, _q_dim = self.y.input_shape()
        (out,) = self.y.output_shape()
        args = [t.to_space("t")]
        if self.inputs is not Noop():
            args.append(u.to_space("u"))
        if p is not None:
            args.append(p.to_space("p"))
        return FunctionSpace("y", args, [out.to_space("y")])


class ParameterMixin(abc.ABC):
    @abc.abstractmethod
    def degrees_of_freedom(self, *interval) -> int:
        pass


@dataclass
class BoundedVariable(ParameterMixin):
    name: str
    lower_bound: float
    upper_bound: float
    guess: float = 0

    def degrees_of_freedom(self, *interval):
        return 1


@dataclass
class PiecewiseConstantVariable(ParameterMixin):
    name: str
    sample_rate: float
    upper_bound: float = np.inf
    lower_bound: float = -np.inf

    def degrees_of_freedom(self, *interval):
        start, end = interval
        return int(np.ceil((end - start) * self.sample_rate))

    def to_solution(self, value):
        return PiecewiseControlSolution(self.name, self.sample_rate, value)


@dataclass
class SpikeVariable(ParameterMixin):
    name: str
    time: float
    upper_bound: float = np.inf
    lower_bound: float = -np.inf

    def degrees_of_freedom(self, *interval):
        return 1

    def to_solution(self, value):
        return SpikeControlSolution(self.name, self.time, value)


@dataclass
class ConstantControlVariable(ParameterMixin):
    name: str
    upper_bound: float = np.inf
    lower_bound: float = -np.inf

    def degrees_of_freedom(self, *interval):
        return 1

    def to_solution(self, value):
        return ConstantControlSolution(self.name, value)


Constant = Union[float, int]
ValueType = Scalar | VectorSpace
ControlLaw = Callable[[Scalar], ValueType]
ControlVariable = (
    ConstantControlVariable | PiecewiseConstantVariable | SpikeVariable
)
ParameterVariable = BoundedVariable | Constant
Solution = Union[
    DynamicalSystem, Callable[[Scalar, ControlLaw, ValueType], Scalar]
]
LossFunction = Callable[[Solution, ControlLaw, ValueType], Scalar]


@dataclass
class TranscriptionOptions:
    minimum_n_intervals: int = 4
    minimum_degree: int = 7
    absolute_tolerance: float = 1e-12
    verbose: bool = False
    optimiser_options: dict = field(default_factory=dict)
    initialise_near_guess: bool = True


@dataclass
class VariationalProblem:
    loss: LossFunction
    system: DynamicalSystem
    t_final: float
    control: Optional[List[ControlVariable]] = None
    parameters: Optional[List[ParameterVariable]] = None
    system_parameter_map: Optional[np.ndarray] = None
    path_constraints: List[InequalityExpression] = field(default_factory=list)
    terminal_constraints: List[InequalityExpression] = field(
        default_factory=list
    )
    transcription_options: TranscriptionOptions = field(
        default_factory=TranscriptionOptions
    )
    backend: Optional[str] = "casadi"

    def __post_init__(self):
        if self.system_parameter_map is not None:
            expected_shape = (
                self.system.parameters.size,
                len(self.parameters),
            )
            assert (
                expected_shape == self.system_parameter_map.shape
            ), f"Parameter map is invalid. Expected an {expected_shape} matrix, but got {self.system_parameter_map.shape}."
        elif (
            self.parameters is not None
            and self.system.parameters.size != len(self.parameters)
        ):
            raise ValueError(
                f"Number of parameters does not match: expected {self.system.parameters.size} but got {len(self.parameters)}. Please provide a parameter map or specify the same number of parameters."
            )

        if self.control is not None:
            assert self.system.inputs is not Noop()

        if not isinstance(self.loss, Function):
            solution_space = self.system.output_as_function_space()
            if self.parameters:
                parameter_space = (
                    VectorSpace("p", len(self.parameters))
                    if self.parameters
                    else None
                )
                solution_space.arguments[-1] = parameter_space
            loss = function(
                arguments=[solution_space, parameter_space],
                implementation=self.loss,
            )
            self.loss = loss

    def __call__(self) -> "VariationalSolution":
        from coker.backends import get_backend_by_name

        backend = get_backend_by_name(self.backend)
        return backend.create_variational_solver(self)


@dataclass
class ConstantControlSolution:
    name: str
    value: float

    def __call__(self, t):
        return self.value


@dataclass
class SpikeControlSolution:
    name: str
    time: float
    value: float
    tolerance: float = 1e-9

    def __call__(self, t):
        return self.value if abs(t - self.time) < self.tolerance else 0.0


@dataclass
class PiecewiseControlSolution:
    name: str
    sample_rate: float
    value: np.ndarray

    def __call__(self, t):
        idx = int(t * self.sample_rate)
        assert idx < len(self.value), f"Time {t} is outside of the interval"
        return self.value[idx]


ControlSolution = Union[
    SpikeControlSolution, PiecewiseControlSolution, ConstantControlSolution
]


class InterpolatingPolyCollection:
    def __init__(self, polys: List["InterpolatingPoly"]):
        self.polys = polys
        self._size = sum(p.size() for p in polys)
        self.intervals = [p.interval for p in polys]

    def size(self):
        return self._size

    def __call__(self, t):
        for i, (start, end) in enumerate(self.intervals):
            if start <= t <= end:
                return self.polys[i](t)
        raise ValueError(f"Value {t} is not in any interval")

    def interval_starts(self):
        for p in self.polys:
            yield p.start_point()

    def interval_ends(self):
        for p in self.polys:
            yield p.end_point()

    def knot_points(self) -> Iterator[Tuple[float, np.ndarray, np.ndarray]]:
        for p in self.polys:
            for point in p.knot_points():
                yield point

    def map(self, func: Callable[[float, np.ndarray], np.ndarray]):
        new_polys = [p.map(func) for p in self.polys]
        return InterpolatingPolyCollection(new_polys)


@dataclass
class VariationalSolution:
    cost: float
    path: InterpolatingPolyCollection
    projectors: Tuple[
        Optional[np.ndarray], Optional[np.ndarray], Optional[np.ndarray]
    ]
    control_solutions: List[ControlSolution]
    parameter_solutions: Dict[str, float]
    parameters: np.ndarray
    output: Callable[
        [float, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray],
        np.ndarray,
    ]

    def as_raw(self) -> np.ndarray:
        points = [
            np.vstack([np.array([t]), x])
            for t, x, _ in self.path.knot_points()
        ]
        return np.hstack(points)

    def state(self, t):
        v = self.path(t)
        proj = self.projectors[0]
        return proj @ v

    def algebraic(self, t):
        if self.projectors[1] is None:
            return None
        return self.projectors[1] @ self.path(t)

    def quadratures(self, t):
        if self.projectors[2] is None:
            return None
        return self.projectors[2] @ self.path(t)

    def control_law(self, t):
        if not self.control_solutions:
            return None
        return np.array([c(t) for c in self.control_solutions])

    def __call__(self, t) -> np.ndarray:
        x = self.state(t)
        q = self.quadratures(t)
        u = self.control_law(t)
        z = self.algebraic(t)
        return self.output(t, x, z, u, self.parameters, q)

    def to_poly(self) -> InterpolatingPolyCollection:

        def f(t, v):
            x = self.projectors[0] @ v
            z = (
                self.projectors[1] @ v
                if self.projectors[1] is not None
                else None
            )
            q = (
                self.projectors[2] @ v
                if self.projectors[2] is not None
                else None
            )
            u = self.control_law(t)
            return self.output(t, x, z, u, self.parameters, q)

        return self.path.map(f)



from typing import List, Union, Optional, NewType, Dict, Tuple
import inspect

from coker import Scalar
from coker.toolkits.system_modelling.modelling.coker_abc import (
    CokerListableSubclasses,
)


def get_component_registry():

    return {
        item.__name__: inspect.getsourcefile(item)
        for item in Block.__subclasses__()
    }


BlockIndex = NewType("BlockIndex", int)
PortIndex = NewType("PortIndex", int)


class LazyPortHandle:
    def __init__(self, component, port_index: PortIndex):
        self.component = component
        self.port_index = port_index

    def resolve(self) -> (BlockIndex, PortIndex):
        parent: BlockContainer = self.component.parent
        assert parent is not None

        index = parent.components.index(self.component)
        return index, self.port_index

    def __repr__(self):
        return f"{self.component}({self.port_index})"


class LazyInputHandle(LazyPortHandle):
    pass


class LazyOutputHandle(LazyPortHandle):
    pass


class CokerModel:
    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def name(self):
        return self.func.__name__


def model(func):
    """Decorator that registers this function as a mode factory"""
    return CokerModel(func)


def is_component_def(obj) -> bool:
    try:
        if issubclass(obj, Block):
            return True
        if issubclass(obj, BlockContainer):
            return True
    except:
        pass
    return False


def is_model(obj) -> bool:
    return isinstance(obj, CokerModel)


class PortForwarder:
    def __init__(self):
        self.ports = []
        self.parent = None

    def set_parent(self, parent: "BlockContainer"):
        assert self.parent is None
        self.parent = parent

    def create(self, signal) -> PortIndex:

        index = len(self.ports)
        self.ports.append(signal)
        return index


class BlockContainer(CokerListableSubclasses):
    def __init__(self, name):
        self.name = name
        """Exposed quantities, forwarded from internal components."""
        self._inputs = PortForwarder()
        self._outputs = PortForwarder()

        self._parameters = []
        self._state = []
        self.parameter_map = {}

        self.labels: Dict[(BlockIndex, PortIndex), str] = {}

        self.components: List[
            Union[Block, "BlockContainer", PortForwarder]
        ] = []
        self.add_models(self._inputs, self._outputs)

        self.parent = None

        """Dictionary of internal connections as input: output"""
        self.connections: Dict[
            (BlockIndex, PortIndex), (BlockIndex, PortIndex)
        ] = {}

    def spec(self):
        return BlockSpec(
            inputs=self._inputs.ports,
            outputs=self._outputs.ports,
            parameters=self._parameters,
            state=self._state,
        )

    def add_models(
        self, *components: Union["Block", "BlockContainer", PortForwarder]
    ):
        for component in components:
            component.set_parent(self)
            self.components.append(component)
            if not (isinstance(component, PortForwarder)):
                spec = component.spec()
                self._state += spec.state if spec.state is not None else []
                self._parameters += (
                    spec.parameters if spec.parameters is not None else []
                )

    def set_parent(self, parent):
        assert (
            self.parent is None
        ), f"{self.name} is already attached to model {self.parent}"
        self.parent = parent

    def add_connections(self, *connections: Tuple):
        for port_in, port_out, *label in connections:
            if isinstance(port_in, LazyOutputHandle) and isinstance(
                port_out, LazyInputHandle
            ):
                port_in, port_out = port_out, port_in
            elif isinstance(port_in, LazyInputHandle) and isinstance(
                port_out, LazyOutputHandle
            ):
                pass
            else:
                raise TypeError("Cannot connect {} to {}", port_in, port_out)
            if len(label) == 0:
                self.add_connection(port_in, port_out)
            elif len(label) == 1:
                (name,) = label
                self.add_connection(port_in, port_out)
                self.labels[port_out.resolve()] = name

    def add_connection(
        self, port_in: LazyInputHandle, port_out: LazyOutputHandle
    ):
        in_port = port_in.resolve()
        out_port = port_out.resolve()
        assert in_port not in self.connections, (
            f"Component input {port_in.component}: {port_in.component.spec().inputs[port_in.port_index]} "
            f"is already in use "
        )
        self.connections[in_port] = out_port

    def add_input(self, signal) -> LazyPortHandle:
        index = self._inputs.create(signal)

        return LazyOutputHandle(self._inputs, index)

    def add_output(self, signal) -> LazyPortHandle:
        index = self._outputs.create(signal)
        return LazyInputHandle(self._outputs, index)


class PortBundle:
    def __init__(self, ports: dict[str, LazyPortHandle]):
        self.ports = ports

    def __getitem__(self, item):
        if isinstance(item, str):
            return self.ports[item]
        if isinstance(item, int):
            values = tuple(self.ports.values())
            return values[item]
        raise AttributeError(item)


class Block(CokerListableSubclasses):
    next_index = 0

    def __init__(self, name: Union[str, None], block_spec: "BlockSpec"):
        self._spec = block_spec
        if name is None:
            name = f"{self.__class__.__name__}_{self.next_index}"
            self.next_index += 1
        self.name = name
        self.parent = None
        self.__generate_io_from_spec()

    def __generate_io_from_spec(self):
        pass

    def spec(self) -> "BlockSpec":
        return self._spec

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def _generate_local_map(self, prefix):
        #    key    value
        #
        pass

    def set_parent(self, parent):
        assert (
            self.parent is None
        ), f"{self.name} is already attached to model {self.parent}"
        self.parent = parent

    @property
    def input(self) -> Union[LazyPortHandle, PortBundle]:
        assert self.spec().inputs, f"Component {self} has no inputs"

        if len(self.spec().inputs) == 1:
            return LazyInputHandle(self, 0)
        else:
            return PortBundle(
                {
                    value.name: LazyInputHandle(self, idx)
                    for idx, value in enumerate(self.spec().inputs)
                }
            )

    @property
    def output(self) -> Union[LazyPortHandle, PortBundle]:
        assert self.spec().outputs, f"Component {self} has no outputs"

        if len(self.spec().outputs) == 1:
            return LazyOutputHandle(self, 0)
        else:
            return PortBundle(
                {
                    value.name: LazyOutputHandle(self, idx)
                    for idx, value in enumerate(self.spec().outputs)
                }
            )


class Signal(Scalar):
    """Real valued analog signal"""

    def __init__(self, name: str, clock_domain=None):
        self.name = name
        self.clock_domain = clock_domain

    def __repr__(self):
        return f"Sig:{self.name}"

    @staticmethod
    def dimension():
        return 1


class Variable:
    def __init__(self, name: str, clock_domain=None):
        self.name = name
        self.clock_domain = clock_domain

    @staticmethod
    def dimension():
        return 1


class VectorVariable(Variable):
    def __init__(self, name: str, dimension: int, clock_domain=None):
        super().__init__(name, clock_domain)
        self._dimension = dimension

    def dimension(self):
        return self._dimension


class Parameter:
    def __init__(self, name: str, value: Optional[float] = None):
        self.name = name
        self.value = value

    def __repr__(self):
        return f"Param:{self.name}"

    def is_constant(self) -> bool:
        return self.value is not None

    @staticmethod
    def dimension():
        return 1


class ConstantParameter:
    def __init__(self, value: float = 0.0):
        self.value = value

    @staticmethod
    def dimension():
        return 1


class Angle(Signal):
    pass


class Position3D(Signal):
    @staticmethod
    def dimension() -> int:
        return 3


class Orientation3D(Signal):
    """Orientation3D in axis-angle representation."""

    @staticmethod
    def dimension() -> 3:
        return 3


class Vector(Parameter):
    pass


class BlockSpec:
    def __init__(
        self,
        inputs: List[Signal] = None,
        state: List[Variable] = None,
        parameters: List[Parameter] = None,
        outputs: List[Signal] = None,
    ):

        self.inputs = inputs
        self.outputs = outputs
        self.state = state
        self.parameters = parameters


ValueType = Union[float, Parameter]



import dataclasses
from typing import Callable, List, Optional
import numpy as np
from coker import VectorSpace, FunctionSpace, function, Scalar, Noop, Function

from .types import DynamicsSpec, DynamicalSystem
from ..algebra import is_scalar
from typing import Tuple


def create_dynamics_from_spec(
    spec: DynamicsSpec, backend="numpy"
) -> DynamicalSystem:

    # just put a dummy value in here so that
    # the shape calculation doesn't spew.

    x0 = function(
        arguments=[
            spec.algebraic,
            spec.inputs,
            spec.parameters,
        ],
        implementation=spec.initial_conditions,
        backend=backend,
    )

    assert (
        len(x0.output) == 2
    ), "Initial conditions must a pair, one for the state and one for the algebraic variables"

    state = x0.output[0]
    algebraic = x0.output[1]

    state_space = VectorSpace("x", state.dim.flat())

    if algebraic is not None:
        assert (
            algebraic.dim == spec.algebraic.dim
        ), "Initial algebraic conditions must have the same dimension as the algebraic variables"

    # Order: t, x, z, u, p
    arguments = [
        Scalar("t"),
        state_space,
        spec.algebraic,
        spec.inputs,
        spec.parameters,
    ]

    xdot = function(arguments, spec.dynamics, backend)

    assert len(xdot.output) == 1, "Dynamics must return a single vector"

    assert (
        xdot.output[0].dim.shape == state.dim.shape
    ), f"Dynamics must return a vector of the same dimension as the state: x0 gave {state.dim} and dynamics gave {xdot.output[0].dim}"

    if spec.algebraic is not None:
        assert (
            spec.constraints is not Noop()
        ), "If algebraic constraints are specified, constraints must be specified"
    elif spec.constraints is not Noop():
        raise ValueError("Constraints specified, but no algebraic variables")

    constraint = (
        function(arguments, spec.constraints, backend)
        if spec.algebraic is not None
        else Noop()
    )

    quadrature = (
        function(arguments, spec.quadratures, backend)
        if spec.quadratures is not Noop()
        else Noop()
    )
    if quadrature is not Noop():
        assert (
            len(quadrature.output) == 1
        ), "Quadratures must be a scalar or vector space"
        q = quadrature.output[0]
        arguments.append(
            VectorSpace("q", q.dim.flat())
            if not q.dim.is_scalar()
            else Scalar("q")
        )
    else:
        arguments.append(None)

    output = function(arguments, spec.outputs, backend)

    return DynamicalSystem(
        spec.inputs, spec.parameters, x0, xdot, constraint, quadrature, output
    )


def create_control_system(
    x0: Callable[[np.ndarray], np.ndarray],
    xdot: Callable[[float, np.ndarray, np.ndarray, np.ndarray], np.ndarray],
    control: FunctionSpace,
    parameters: Optional[Scalar | VectorSpace] = None,
    output: Callable[
        [Scalar, np.ndarray, np.ndarray, np.ndarray], np.ndarray
    ] = None,
    backend: str = "numpy",
    p_init: np.ndarray = None,
    u_init: Callable[[float], np.ndarray] = None,
) -> DynamicalSystem:

    if isinstance(x0, (list, tuple, int, float)):
        x0 = np.array(x0)
    if isinstance(x0, np.ndarray):
        x0_func = lambda z, u, p: (x0, None)
    else:
        x0_func = lambda z, u, p: (x0(p), None)

    if p_init is not None:
        assert (
            parameters is not None
        ), "Parameters must be specified if p_init is specified"
        if isinstance(parameters, Scalar):
            assert isinstance(p_init, (float, int)) or p_init.shape == (
                1,
            ), "p_init must be a scalar or a 1D array"
            p_init = np.array([p_init])
        else:
            assert p_init.shape == (
                parameters.dimension,
            ), "p_init must be a 1D array of the same size as the parameters"
    else:
        p_init = None if parameters is None else np.zeros(parameters.dimension)

    assert len(control.output) == 1, "Control must return a single vector"
    (u_dim,) = control.output_dimensions()
    if u_init is not None:
        assert callable(u_init), "u_init must be a callable"
        u0 = u_init(0)
        assert (
            u0.shape == u_dim.shape
        ), f"u0 must have the same shape as the control output; u0 is {u0} and control output is {u_dim}"
    else:
        u0 = np.zeros(u_dim.shape)

    x0_eval, _z0 = x0_func(None, u0, p_init)
    dot_x_eval = xdot(0, x0_eval, u0, p_init)

    assert (
        is_scalar(x0_eval) and is_scalar(dot_x_eval)
    ) or x0_eval.shape == dot_x_eval.shape, f"x0 and xdot must have the same shape; x0 is {x0_eval} and xdot is {dot_x_eval}"

    if output is None:
        output_func = lambda t, x, z, u, p, q: x
    else:
        y_eval = output(0, x0_eval, u0, p_init)
        assert y_eval is not None, "Output function must return a value"
        output_func = lambda t, x, z, u, p, q: output(t, x, u(t), p)

    dynamics = lambda t, x, z, u, p: xdot(t, x, u(t), p)

    spec = DynamicsSpec(
        inputs=control,
        parameters=parameters,
        algebraic=None,
        initial_conditions=x0_func,
        dynamics=dynamics,
        constraints=Noop(),
        outputs=output_func,
        quadratures=Noop(),
    )

    return create_dynamics_from_spec(spec, backend=backend)


def create_autonomous_ode(
    x0: Callable[[List[np.ndarray]], np.ndarray],
    xdot: Callable[[np.ndarray, np.ndarray, np.ndarray], np.ndarray],
    parameters: Optional[Scalar | VectorSpace] = None,
    output: Callable[[np.ndarray, np.ndarray], np.ndarray] = None,
    backend: str = "numpy",
    p_init: np.ndarray = None,
) -> DynamicalSystem:

    # case 1,
    # - x0 is an array

    if isinstance(x0, (list, tuple)):
        x0 = np.array(x0)

    if isinstance(x0, (np.ndarray, int, float)):
        x0_func = lambda z, u, p: (x0, None)
    else:
        x0_func = lambda z, u, p: (x0(p), None)

    if p_init is not None:
        assert (
            parameters is not None
        ), "Parameters must be specified if p_init is specified"
        if isinstance(parameters, Scalar):
            assert isinstance(p_init, (float, int)) or p_init.shape == (
                1,
            ), "p_init must be a scalar or a 1D array"
            p_init = np.array([p_init])
        else:
            assert p_init.shape == (
                parameters.dimension,
            ), "p_init must be a 1D array of the same size as the parameters"
    else:
        p_init = None if parameters is None else np.zeros(parameters.dimension)

    x0_eval, _z0 = x0_func(None, None, p_init)
    dot_x_eval = xdot(x0_eval, p_init)

    assert (
        is_scalar(x0_eval) and is_scalar(dot_x_eval)
    ) or x0_eval.shape == dot_x_eval.shape, f"x0 and xdot must have the same shape; x0 is {x0_eval} and xdot is {dot_x_eval}"

    if output is None:
        if is_scalar(x0_eval):
            output_func = lambda t, x, z, u, p, q: x[0]
        else:
            output_func = lambda t, x, z, u, p, q: x

    else:
        y_eval = output(x0_eval, p_init)
        assert y_eval is not None, "Output function must return a value"
        output_func = (lambda t, x, z, u, p, q: output(x, p),)

    dynamics = lambda t, x, z, u, p: xdot(x, p)

    spec = DynamicsSpec(
        inputs=Noop(),
        parameters=parameters,
        algebraic=None,
        initial_conditions=x0_func,
        dynamics=dynamics,
        constraints=Noop(),
        outputs=output_func,
        quadratures=Noop(),
    )

    return create_dynamics_from_spec(spec, backend=backend)


class CompositionOperator:

    def __init__(self, *spaces: Scalar | VectorSpace | None):
        self.spaces = spaces

    def offsets(self):
        total = 0
        for space in self.spaces:
            yield total
            if space is not None:
                total += space.size

    def sizes(self):
        for space in self.spaces:
            if space is None:
                yield 0
            else:
                yield space.size

    def dim(self) -> Scalar | VectorSpace | None:

        dim = sum(self.sizes())
        return VectorSpace(f"composition", (dim,)) if dim else None

    def inverse(self, ab) -> List:
        next_slice = ab
        output = []
        for space in self.spaces:
            if space is None:
                output.append(None)
            else:
                dim = space.size
                output.append(next_slice[:dim])
                next_slice = (
                    next_slice[dim:] if dim < next_slice.shape[0] else []
                )
        return output

    def __call__(self, *args):
        output = [
            a for a, space in zip(args, self.spaces) if space is not None
        ]
        if not output:
            return None

        return np.concatenate(output)

    @staticmethod
    def from_dimensions(name: str, *dims) -> "CompositionOperator":
        spaces = [
            VectorSpace(f"{name}_{i}", dim) if dim else None
            for i, dim in enumerate(dims)
        ]
        return CompositionOperator(*spaces)

    def as_matrices(self):
        offsets = list(self.offsets())
        sizes = list(self.sizes())

        matrices = []
        for i, (offset, size) in enumerate(zip(offsets, sizes)):
            matrix = np.zeros((size, sum(self.sizes())), dtype=float)

            matrix[:, offset : offset + size] = np.eye(size)
            matrices.append(matrix)
        return matrices


@dataclasses.dataclass
class ProjectionSet:
    parameters: CompositionOperator
    state: CompositionOperator
    algebraic: CompositionOperator
    quadratures: CompositionOperator
    outputs: CompositionOperator
    controls: CompositionOperator


def direct_sum(
    *systems: DynamicalSystem, backend=None
) -> Tuple[DynamicalSystem, ProjectionSet]:

    backend = backend or systems[0].backend()

    proj_p = CompositionOperator(*[system.parameters for system in systems])
    x_dim, z_dim, q_dim = zip(
        *[system.get_state_dimensions() for system in systems]
    )
    y_dim = [system.y.output_shape()[0] for system in systems]

    proj_x = CompositionOperator.from_dimensions("x", *x_dim)
    proj_z = CompositionOperator.from_dimensions("z", *z_dim)
    proj_q = CompositionOperator.from_dimensions("q", *q_dim)

    proj_y = CompositionOperator.from_dimensions("y", *y_dim)

    u_range = [
        (
            system.inputs.output_dimensions()[0]
            if system.inputs is not Noop()
            else None
        )
        for system in systems
    ]

    proj_u = CompositionOperator.from_dimensions(f"u", *u_range)
    if proj_u.dim() is None:
        u_space = Noop()
    else:
        u_space = FunctionSpace(
            proj_u.dim().name, [Scalar("t")], [proj_u.dim()]
        )

    def x0_impl(t, u_outer, p_outer):
        p_inner = proj_p.inverse(p_outer)
        u_inner = [
            lambda t: proj_u.inverse(u_outer(t))[i]
            for i, _ in enumerate(proj_u.spaces)
        ]
        x0_inner, z0_inner = zip(
            *[
                system.x0.call_inline(t, u_i, p_i)
                for system, u_i, p_i in zip(systems, u_inner, p_inner)
            ]
        )
        x0_ab = proj_x(*x0_inner)
        z0_ab = proj_z(*z0_inner)
        return x0_ab, z0_ab

    x0 = function(
        [Scalar("t"), u_space, proj_p.dim()], x0_impl, backend=backend
    )

    def dxdt_impl(t, x_outer, z_outer, u_outer, p_outer):
        p = proj_p.inverse(p_outer)
        x = proj_x.inverse(x_outer)
        z = proj_z.inverse(z_outer)

        dx_inner = [
            system.dxdt.call_inline(
                t,
                x[i],
                z[i],
                lambda t_i: proj_u.inverse(u_outer(t_i))[i],
                p[i],
            )
            for i, system in enumerate(systems)
        ]

        return proj_x(*dx_inner)

    args = [Scalar("t"), proj_x.dim(), proj_z.dim(), u_space, proj_p.dim()]
    dx = function(args, dxdt_impl, backend=backend)
    if proj_z.dim() is not None:

        def g_impl(t, x_outer, z_outer, u_outer, p_outer):
            p = proj_p.inverse(p_outer)
            x = proj_x.inverse(x_outer)
            z = proj_z.inverse(z_outer)
            g_inner = [
                system.g.call_inline(
                    t,
                    x[i],
                    z[i],
                    lambda t_i: proj_u.inverse(u_outer(t_i))[i],
                    p[i],
                )
                for i, system in enumerate(systems)
            ]
            return proj_z(*g_inner)

        g = function(args, g_impl, backend=backend)
    else:
        g = Noop()

    if proj_q.dim() is not None:

        def dqdt_impl(t, x_outer, z_outer, u_outer, p_outer):
            p = proj_p.inverse(p_outer)
            x = proj_x.inverse(x_outer)
            z = proj_z.inverse(z_outer)
            dq_inner = [
                system.dqdt.call_inline(
                    t,
                    x[i],
                    z[i],
                    lambda t_i: proj_u.inverse(u_outer(t_i))[i],
                    p[i],
                )
                for i, system in enumerate(systems)
            ]
            return proj_q(*dq_inner)

        dqdt = function(args, dqdt_impl, backend=backend)
    else:
        dqdt = Noop()

    def y_impl(t, x_outer, z_outer, u_outer, p_outer, q_outer):
        p = proj_p.inverse(p_outer)
        x = proj_x.inverse(x_outer)
        z = proj_z.inverse(z_outer)
        q = proj_q.inverse(q_outer)
        y_inner = [
            system.y.call_inline(
                t,
                x[i],
                z[i],
                lambda t_i: proj_u.inverse(u_outer(t_i))[i],
                p[i],
                q[i],
            )
            for i, system in enumerate(systems)
        ]

        return proj_y(*y_inner)

    y = function(args + [proj_q.dim()], y_impl, backend=backend)

    system = DynamicalSystem(
        inputs=u_space,
        parameters=proj_p.dim(),
        x0=x0,
        dxdt=dx,
        g=g,
        dqdt=dqdt,
        y=y,
    )
    projections = ProjectionSet(
        state=proj_x,
        algebraic=proj_z,
        quadratures=proj_q,
        outputs=proj_y,
        controls=proj_u,
        parameters=proj_p,
    )

    return system, projections



from types import NotImplementedType
from typing import Optional
import numpy as np
from coker.toolkits.spatial.types import Vec3, Scalar
from coker.toolkits.spatial.unit_quaternion import UnitQuaternion
from coker.algebra.kernel import normalise

SE3_BASIS = np.array(
    [
        [[0, 0, 0], [0, 0, -1], [0, 1, 0]],
        [[0, 0, 1], [0, 0, 0], [-1, 0, 0]],
        [[0, -1, 0], [1, 0, 0], [0, 0, 0]],
    ],
)

e_x = np.array([1, 0, 0])
e_y = np.array([0, 1, 0])
e_z = np.array([0, 0, 1])


def hat(u: Vec3):
    result = SE3_BASIS @ u
    return result


class Rotation3:
    def __init__(self, axis, angle):
        try:
            if np.allclose(axis, np.zeros(3)):
                axis = e_z
                angle = 0
        except NotImplementedError:
            pass
        self.axis = axis
        self.angle = angle

    @staticmethod
    def cast(other) -> "Rotation3":

        if isinstance(other, Rotation3):
            return other

        if isinstance(other, UnitQuaternion):
            return Rotation3.from_quaterion(other)
        if isinstance(other, Vec3):
            return Rotation3.from_vector(other)
        if hasattr(other, "shape") and other.shape == (3,):
            return Rotation3.from_vector(other)

        raise NotImplementedError(f"Cannot cast {other} to a rotation matrix")

    def as_vector(self) -> np.ndarray:
        return self.axis * self.angle

    def __repr__(self):
        return f"Rotation3({repr(self.axis)}, {repr(self.angle)})"

    @staticmethod
    def zero():
        return Rotation3(
            np.array(
                e_z.copy(),
            ),
            angle=0,
        )

    @staticmethod
    def from_vector(vector):
        array = np.array(
            vector,
        )
        angle = np.linalg.norm(array)
        if angle == 0:
            return Rotation3.zero()
        axis = array / angle
        return Rotation3(axis, angle)

    def inverse(self):
        return Rotation3(self.axis, -self.angle)

    def as_quaternion(self):
        return UnitQuaternion.from_axis_angle(self.axis, self.angle)

    @staticmethod
    def from_quaterion(q: UnitQuaternion):
        try:
            if isinstance(q.q_0, (float, int)) and q.q_0 == 1:
                return Rotation3.zero()
        except NotImplementedError:
            pass
        u, r = normalise(q.v)
        theta = 2 * np.arctan2(r, q.q_0)
        #        theta = 2 * np.arccos(q.q_0)
        #        r = np.sqrt(1 - q.q_0 * q.q_0)  # sin(theta/2)
        #        u = q.v / r
        return Rotation3(axis=u, angle=theta)

    def __mul__(self, other: "Rotation3"):
        q_1 = self.as_quaternion()
        q_2 = other.as_quaternion()

        q_21 = q_2 * q_1

        return Rotation3.from_quaterion(q_21)

    def apply(self, other: np.ndarray):
        return self.as_quaternion().conjugate(other)

    def __matmul__(self, other: np.ndarray):
        assert other.shape == (3,)
        return self.apply(other)

    def __eq__(self, other):
        return (
            isinstance(other, Rotation3)
            and (self.axis == other.axis).all()
            and self.angle == other.angle
        )

    def as_matrix(self):
        k = hat(self.axis)
        s = np.sin(self.angle)
        c = np.cos(self.angle)

        return np.eye(3) + s * k + (1 - c) * (k @ k)


class Isometry3:
    def __init__(
        self,
        rotation: Optional[Rotation3] = None,
        translation: Optional[Vec3] = None,
    ):

        self.rotation = (
            Rotation3.zero()
            if rotation is None
            else (Rotation3.cast(rotation))
        )

        self.translation = (
            np.array(
                [0, 0, 0],
            )
            if translation is None
            else translation
        )

    def __repr__(self):
        return (
            f"Isometry3(r={repr(self.rotation)}, t={repr(self.translation)})"
        )

    def __eq__(self, other):
        return (
            isinstance(other, Isometry3)
            and self.rotation == other.rotation
            and (self.translation == other.translation).all()
        )

    @staticmethod
    def identity():
        return Isometry3(
            Rotation3.zero(),
            translation=np.array(
                [0, 0, 0],
            ),
        )

    def __matmul__(self, other):
        if isinstance(other, Isometry3):
            rotation = other.rotation * self.rotation
            translation = (
                self.rotation.as_quaternion().conjugate(other.translation)
                + self.translation
            )
            return Isometry3(rotation, translation)
        elif isinstance(other, np.ndarray) and other.shape == (3,):
            return (
                self.rotation.as_quaternion().conjugate(other)
                + self.translation
            )
        elif isinstance(other, np.ndarray) and other.shape == (4, 1):
            result = np.reshape(
                self.rotation.as_quaternion().conjugate(other[0:3, 0])
                + self.translation,
                shape=(3, 1),
            )
            return np.concatenate([result, other[3:4, 0:1]], axis=0)

        raise NotImplementedError(f"Isometry can't tranfrom {other.__class__}")

    def apply(self, other):
        assert other.shape == (3,)
        return (
            self.rotation.as_quaternion().conjugate(other) + self.translation
        )

    def transpose(self):
        q = self.rotation.inverse()
        return Isometry3(q, -q.as_quaternion().conjugate(self.translation))

    def as_matrix(self):
        r = self.rotation.as_matrix()
        p = np.reshape(self.translation, shape=(3, 1))
        o = np.array(
            [[0, 0, 0, 1.0]],
        )

        result = np.concatenate((r, p), axis=1)

        return np.concatenate((result, o))

    def inverse(self):
        r = self.rotation.inverse()
        p = r.as_matrix() @ self.translation
        return Isometry3(rotation=r, translation=-p)

    def as_vector(self) -> np.ndarray:

        p = np.reshape(self.translation, shape=(3, 1))
        q = np.reshape(self.rotation.as_vector(), shape=(3, 1))
        return np.concatenate((q, p), axis=0)

    @staticmethod
    def from_vector(vector: np.ndarray):

        q = vector[0:3]
        p = vector[3:6]
        r = Rotation3.from_vector(q)
        return Isometry3(rotation=r, translation=p)


class Screw:
    __slots__ = ("rotation", "translation", "magnitude")

    def __init__(
        self,
        rotation: Optional[Vec3] = None,
        translation: Optional[Vec3] = None,
        magnitude: float = 1,
    ):
        self.rotation = (
            rotation if rotation is not None else np.array([0, 0, 0])
        )
        self.translation = (
            translation if translation is not None else np.array([0, 0, 0])
        )
        self.magnitude = magnitude

    def __repr__(self):
        l = self.to_array().tolist()
        return repr(l)

    @staticmethod
    def w_z():
        return Screw(rotation=e_z.copy())

    @staticmethod
    def w_x():
        return Screw(rotation=e_x.copy())

    @staticmethod
    def w_y():
        return Screw(rotation=e_y.copy())

    @staticmethod
    def e_z():
        return Screw(translation=e_z.copy())

    @staticmethod
    def e_x():
        return Screw(translation=e_x.copy())

    @staticmethod
    def e_y():
        return Screw(translation=e_y.copy())

    def __neg__(self):
        return Screw(rotation=-self.rotation, translation=-self.translation)

    @staticmethod
    def from_tuple(*values):
        rotation = np.array(values[0:3])
        translation = np.array(values[3:6])
        if all(v == 0 for v in values[0:3]):
            return Screw(rotation, translation, 1)

        mag = np.linalg.norm(rotation).astype(float).flatten()[0]

        return Screw(rotation / mag, translation / mag, mag)

    def to_array(self) -> np.ndarray:
        return (
            np.concatenate([self.rotation, self.translation]) * self.magnitude
        )

    @staticmethod
    def from_array(array):
        assert array.shape == (6,)
        rotation = array[0:3]
        translation = array[3:6]
        try:
            if (rotation == 0.0).all():
                return Screw(rotation, translation, 1)
            else:
                mag = np.linalg.norm(rotation)
        except TypeError:
            mag = 1

        return Screw(rotation / mag, translation / mag, mag)

    @staticmethod
    def zero():
        return Screw(
            rotation=np.array(
                [1, 0, 0],
            ),
            translation=np.array(
                [0, 0, 0],
            ),
            magnitude=0,
        )

    def __mul__(self, other: Scalar):
        return Screw(self.rotation, self.translation, self.magnitude * other)

    @property
    def axis(self) -> (Vec3, Vec3):
        """
        Returns:
            Tuple[Vec3, Vec3]: Point and direction of axis.

        """
        q = np.cross(self.rotation, self.translation)

        if (self.rotation == 0).all():
            return q, self.translation

        return q, self.rotation

    @property
    def pitch(self):
        return self.rotation.dot(self.translation) * self.magnitude

    def exp(self, angle=1) -> Isometry3:
        if self.magnitude != 0:
            alpha = self.magnitude * angle
        else:
            alpha = angle

        if (
            isinstance(self.rotation, np.ndarray)
            and (self.rotation == 0).all()
        ):
            return Isometry3(
                rotation=Rotation3.zero(), translation=self.translation * alpha
            )

        rotation = Rotation3(axis=self.rotation, angle=alpha)

        rot_dot_t = np.dot(self.rotation, self.translation)
        rot_cross_t = np.cross(self.rotation, self.translation)

        t1 = -rotation.apply(rot_cross_t)
        t2 = alpha * self.rotation
        t3 = rot_dot_t * t2

        translation = rot_cross_t + t1 + t3

        return Isometry3(rotation=rotation, translation=translation)

    def as_matrix(self):
        w_hat = hat(self.rotation)
        return np.concatenate(
            np.concatenate([w_hat, self.translation]),
            np.array([[0, 0, 0, 1]]),
            axis=1,
        )

    def __eq__(self, other):
        return (
            isinstance(other, Screw)
            and (self.rotation == other.rotation).all()
            and (self.translation == other.translation).all()
            and self.magnitude == other.magnitude
        )


class SE3Adjoint:
    def __init__(self, transform: Isometry3):
        self.transform = transform

    def apply(self, zeta: Screw) -> Screw:

        q = self.transform.rotation.as_quaternion()
        p = self.transform.translation

        rotation = q.conjugate(zeta.rotation)
        translation = hat(p) @ rotation + q.conjugate(zeta.translation)

        return Screw(
            rotation=rotation,
            translation=translation,
            magnitude=zeta.magnitude,
        )

    def transpose(self):
        return SE3CoAdjoint(self.transform)

    def __matmul__(self, other):
        if isinstance(other, Screw):
            return self.apply(other)
        if isinstance(other, SE3Adjoint):
            t = self.transform @ other.transform
            return SE3Adjoint(t)
        if isinstance(other, np.ndarray):
            return self.as_matrix() @ other

        raise NotImplementedError(f"Can't matmul by {other.__class__}")

    def __repr__(self):
        return f"Adj[{repr(self.transform)}]"

    def __rmatmul__(self, other):
        if isinstance(other, SE3Adjoint):
            t = other.transform @ self.transform
            return SE3Adjoint(t)
        raise NotImplementedError(f"Can't rmatmul by {other.__class__}")

    def __call__(self, arg: Screw) -> Screw:
        assert isinstance(arg, Screw)
        return self.apply(arg)

    def inverse(self):
        return SE3Adjoint(self.transform.inverse())

    def as_matrix(self):
        r = self.transform.rotation.as_matrix()
        p_hat = hat(self.transform.translation)
        row_1 = np.concatenate(
            [
                r,
                np.zeros(
                    (3, 3),
                ),
            ],
            axis=1,
        )
        row_2 = np.concatenate([p_hat @ r, r], axis=1)
        return np.concatenate([row_1, row_2], axis=0)


class SE3CoAdjoint:
    def __init__(self, transform: Isometry3):
        self.transform = transform

    def as_matrix(self):
        r = self.transform.rotation.as_matrix().T
        p_hat = hat(self.transform.translation)
        row_1 = np.concatenate([r, -r @ p_hat], axis=1)
        row_2 = np.concatenate(
            [
                np.zeros(
                    (3, 3),
                ),
                r,
            ],
            axis=1,
        )
        return np.concatenate([row_1, row_2], axis=0)

    def __repr__(self):
        return f"Adj^*[{repr(self.transform)}]"

    def apply(self, zeta: Screw) -> Screw:
        q_inv = self.transform.rotation.as_quaternion().inverse()
        p = self.transform.translation

        translation = q_inv.conjugate(zeta.translation)
        rotation = -q_inv.conjugate(
            hat(p) @ zeta.translation
        ) + q_inv.conjugate(zeta.rotation)

        return Screw(
            rotation=rotation,
            translation=translation,
            magnitude=zeta.magnitude,
        )

    def __matmul__(self, other):
        if isinstance(other, Screw):
            return self.apply(other)
        raise NotImplementedError(f"Can't matmul by type {other.__class__}")

    def __call__(self, arg: Screw) -> Screw:
        assert isinstance(arg, Screw)
        return self.apply(arg)

    def transpose(self):
        return SE3Adjoint(self.transform)


class se3Adjoint:
    def __init__(self, vector: Screw):
        self.vector = vector

    def apply(self, other: Screw) -> Screw:
        w = np.cross(self.vector.rotation, other.rotation)
        v = np.cross(self.vector.translation, other.rotation) - np.cross(
            self.vector.rotation, other.translation
        )

        m = self.vector.magnitude * other.magnitude

        return Screw(translation=v, rotation=w, magnitude=m)

    def transpose(self) -> "se3CoAdjoint":
        return se3CoAdjoint(self.vector)

    def __matmul__(self, other: Screw):
        assert isinstance(other, Screw)

        return self.apply(other)


class se3CoAdjoint:
    def __init__(self, vector: Screw):
        self.vector = vector

    def __matmul__(self, other: Screw):
        assert isinstance(other, Screw)

        return self.apply(other)

    def apply(self, other: Screw) -> Screw:
        w = np.cross(self.vector.rotation, other.rotation) + np.cross(
            self.vector.translation, other.translation
        )

        v = np.cross(self.vector.rotation, other.translation)

        m = self.vector.magnitude * other.magnitude

        return Screw(translation=-v, rotation=-w, angle=m)

    def transpose(self) -> se3Adjoint:
        return se3Adjoint(self.vector)


def se3_bracket(left: Screw, right: Screw) -> Screw:
    w = np.cross(left.rotation, right.rotation)
    v = np.cross(left.rotation, right.translation) - np.cross(
        right.rotation, left.translation
    )

    return Screw(
        translation=v, rotation=w, angle=right.magnitude * left.magnitude
    )



from typing import Type, Tuple, List, Any
from functools import reduce
from operator import mul
from enum import Enum
import numpy as np
import scipy.sparse.csc
import sympy as sp
import scipy as sy
import scipy as scp

from coker.algebra import Dimension, OP
from coker.algebra.kernel import Tracer, VectorSpace, Noop
from coker.algebra.ops import ConcatenateOP, ReshapeOP, NormOP

from coker.backends.backend import Backend, ArrayLike, SolverParameters
from coker.backends.evaluator import evaluate_inner


class Solver(Enum):
    RK45 = "RK45"
    LSODA = "LSODA"
    Radau = "Radau"
    BDF = "BDF"


class NumpySolverParameters(SolverParameters):
    def __init__(self, solver: Solver = Solver.RK45):
        self.solver = solver


def to_array(value, shape):

    if isinstance(value, np.ndarray) and value.shape == shape:
        return value

    raise NotImplementedError


scalar_types = (
    np.float32,
    np.float64,
    np.int32,
    np.int64,
    float,
    complex,
    int,
    bool,
    np.bool_,
)


def is_scalar_symbol(v):
    if isinstance(v, sp.MatrixSymbol):
        return v.shape == (1, 1)
    return isinstance(v, (sp.Symbol, sp.Expr))


def div(num, den):
    if isinstance(den, Tracer):
        return num / den
    try:
        if den == 0:
            if isinstance(num, np.ndarray) and (num == 0).all():
                return num
            elif num == 0:
                return num
            else:
                raise ZeroDivisionError(num, den)
    except ValueError:
        pass

    return np.divide(num, den)


impls = {
    OP.ADD: np.add,
    OP.SUB: np.subtract,
    OP.MUL: np.multiply,
    OP.DIV: div,
    OP.MATMUL: np.matmul,
    OP.SIN: np.sin,
    OP.COS: np.cos,
    OP.TAN: np.tan,
    OP.EXP: np.exp,
    OP.PWR: np.power,
    OP.INT_PWR: np.power,
    OP.ARCCOS: np.arccos,
    OP.ARCSIN: np.arcsin,
    OP.DOT: np.dot,
    OP.CROSS: np.cross,
    OP.TRANSPOSE: np.transpose,
    OP.NEG: np.negative,
    OP.SQRT: np.sqrt,
    OP.ABS: np.abs,
    OP.ARCTAN2: np.arctan2,
    OP.EQUAL: np.equal,
    OP.LESS_EQUAL: np.less_equal,
    OP.LESS_THAN: np.less,
    OP.CASE: lambda cond, t, f: t if cond else f,
    OP.EVALUATE: lambda op, *args: op(*args),
    OP.LOG: np.log,
}

parameterised_impls = {
    ConcatenateOP: lambda op, *x: np.concatenate(x, axis=op.axis),
    ReshapeOP: lambda op, x: np.reshape(x, shape=op.newshape),
    NormOP: lambda op, x: np.linalg.norm(x, ord=op.ord),
}


def call_parameterised_op(op, *args):
    kls = op.__class__

    result = parameterised_impls[kls](op, *args)

    return result


def jacobian(f: sp.Expr, x):
    result = sp.Matrix([f]).jacobian(x)
    return result


def hessian(f: sp.Expr, x):
    h = sp.hessian(f, x)
    return h


def reshape_sympy_matrix(arg, shape):

    if len(shape) == 1:
        out_shape = (shape[0], 1)
    else:
        assert len(shape) == 2
        out_shape = shape

    if arg.shape == (*out_shape, 1):
        return arg[:, :, 0]

    old_cols = 1 if len(arg.shape) == 1 else arg.shape[1]

    def lookup(i, j):
        offset = i * out_shape[1] + j
        old_row = offset // old_cols
        old_col = offset % old_cols
        value = arg[old_row, old_col]
        return value

    return sp.Matrix(*out_shape, lookup)


def reshape(arg, dim):
    if dim.is_scalar():
        if isinstance(arg, scalar_types) or is_scalar_symbol(arg):
            return arg
        else:
            try:
                (inner,) = arg
            except ValueError as ex:
                raise TypeError(f"Expecting a scalar, got {arg}") from ex
            except TypeError as ex:
                raise TypeError(f"Expecting a scalar, got {arg}") from ex
            return reshape(inner, dim)
    elif isinstance(
        arg, (sp.Matrix, sp.Array, sp.MatrixSlice, sp.ImmutableMatrix)
    ):
        if arg.shape == dim.dim:
            return arg
        return reshape_sympy_matrix(arg, dim.dim)
    elif isinstance(arg, np.ndarray):
        return np.reshape(arg, dim.dim)
    elif scp.sparse.issparse(arg):
        return np.reshape(arg.toarray(), dim.dim)
    elif isinstance(arg, (float, int)):
        return np.array([arg]).reshape(dim.dim)
    elif arg is None:
        return arg
    raise NotImplementedError(f"Dont know how to reshape {arg}")


class NumpyBackend(Backend):
    def __init__(self, *args, **kwargs):
        super(NumpyBackend, self).__init__(*args, **kwargs)

    def native_types(self) -> List[Any]:
        return [
            np.ndarray,
            np.int32,
            np.int64,
            np.float64,
            np.float32,
            float,
            complex,
            int,
        ]

    def to_numpy_array(self, array) -> ArrayLike:
        return array

    def to_backend_array(self, array: ArrayLike):
        return array

    def reshape(self, arg, dim: Dimension):
        return reshape(arg, dim)
        raise NotImplementedError(
            f"Don't know how to resize {arg.__class__.__name__}"
        )

    def lower(self, function):
        from coker.backends.evaluator import _build_plan, _cast_outputs

        plan = _build_plan(function.tape, self)
        tape = function.tape
        outputs = function.output
        backend = self

        def compiled(inputs):
            workspace = plan.execute(inputs, backend)
            return _cast_outputs(outputs, tape, workspace, backend)

        return compiled

    def resolve_fn(self, op):
        if op in impls:
            return impls[op]
        if isinstance(op, tuple(parameterised_impls.keys())):
            kls = op.__class__
            _op = op
            return lambda *args: parameterised_impls[kls](_op, *args)
        raise NotImplementedError(f"{op} is not implemented")

    def resolve_post_fn(self, dim):
        # For scalar outputs, reshape extracts the Python scalar from the array.
        # Tracers must pass through unchanged (they appear during function composition tracing).
        # For non-scalar outputs, numpy ops already produce the correct shape.
        if dim.is_scalar():
            _dim = dim

            def scalar_post(v):
                if isinstance(v, Tracer):
                    return v
                return reshape(v, _dim)

            return scalar_post
        return lambda v: v

    def call(self, op, *args) -> ArrayLike:

        if op in impls:
            return impls[op](*args)

        if isinstance(op, tuple(parameterised_impls.keys())):
            return call_parameterised_op(op, *args)

        raise NotImplementedError(f"{op} is not implemented")

    def evaluate_integrals(
        self,
        functions,
        initial_conditions,
        end_point: float,
        inputs,
        solver_parameters=None,
    ):

        dxdt, constraint, dqdt = functions
        x0, z0, q0 = initial_conditions
        u, p = inputs

        if constraint is not Noop():
            raise NotImplementedError(
                "Integrators with constraints are not implemented"
            )

        if not isinstance(x0, np.ndarray):
            x0 = np.array([x0])

        if isinstance(end_point, (float, int)):
            if end_point == 0.0:
                return x0, z0, q0
            else:
                t_eval = [end_point]
                t_span = (0, end_point)
        else:
            t_eval = end_point
            t_span = (0, end_point[-1])

        if dqdt is Noop():
            y0 = x0
            f = lambda t, x: dxdt(t, x, None, u, p)

        else:
            y0 = (np.concatenate([x0, q0]),)
            f = lambda t, x: np.concatenate(
                [dxdt(t, x, None, u, p), dqdt(t, x, None, u, p)]
            )

        if isinstance(solver_parameters, NumpySolverParameters):
            method = solver_parameters.solver.value
        else:
            method = Solver.RK45.value

        sol = scp.integrate.solve_ivp(
            f, t_span, y0, method=method, t_eval=t_eval
        )

        x_out = (
            sol.y[: x0.shape[0], -1]
            if not isinstance(end_point, np.ndarray)
            else sol.y[: x0.shape[0], :]
        )

        if dqdt is None:
            q_out = None
        else:
            q_out = (
                sol.y[x0.shape[0] :, -1]
                if not isinstance(end_point, np.ndarray)
                else sol.y[x0.shape[0] :, :]
            )

        return x_out, None, q_out

    def build_optimisation_problem(
        self,
        cost: Tracer,  # cost
        constraints: List[Tracer],
        arguments: List[Tracer],
        outputs: List[Tracer],
    ):

        impl = None
        tape = cost.tape

        assert all(c.tape == tape for c in constraints)
        assert all(a.tape == tape for a in arguments)
        assert all(o.tape == tape for o in outputs)

        # For each variable
        #
        # 2. replace each 'variable' with a projection P, such that
        # V_i = P_iX + \sum_j R_jA_j
        # where A_j is the jth argument

        # construct Q_0, P_0, R_0 and N(X)
        # c_i = X^T Q_i X + P_i X + R_i + N_i(Z)

        arg_indexes = {a.index for a in arguments}
        decision_variables = [
            i for i in tape.input_indicies if i not in arg_indexes
        ]

        arg_symbols = []
        for i, a in enumerate(arguments):
            shape = (a.shape[0], 1) if len(a.shape) == 1 else a.shape
            assert len(shape) == 2
            arg_symbols.append(sp.MatrixSymbol(f"p_{i}", *shape))

        n = 0
        mappings = []
        for index in decision_variables:
            dim: Dimension = tape.dim[index]
            if dim.is_scalar():
                flat_dim = 1
            else:
                flat_dim = reduce(mul, dim)

            mappings.append((index, dim, (n, flat_dim + n)))
            n += flat_dim

        x = sp.Array([sp.symbols(f"x_{i}") for i in range(n)])
        for idx, dim, (start, stop) in mappings:
            tape.substitute(idx, (OP.VALUE, x[start:stop]))

        workspace = {}
        (cost,) = evaluate_inner(tape, arguments, [cost], self, workspace)

        problem_args = [x, *arg_symbols]
        cost_f = sp.lambdify(problem_args, cost)

        cost_jac = lambda a: sp.lambdify(problem_args, jacobian(cost, x))(
            a
        ).reshape(x.shape)
        cost_hess = sp.lambdify(problem_args, hessian(cost, x))

        out_constriants = []
        for i, constraint in enumerate(constraints):
            (c,) = evaluate_inner(
                tape,
                arguments,
                [constraint.as_halfplane_bound()],
                self,
                workspace,
            )
            c_func = sp.lambdify(problem_args, c)
            c_jac = jacobian(c, problem_args)

            if not c_jac.free_symbols:
                c_0 = c_func(*[np.zeros_like(x_i) for x_i in problem_args])
                this_constraint = sy.optimize.LinearConstraint(
                    c_jac.evalf(), -c_0, np.inf
                )
            else:
                this_constraint = sy.optimize.NonlinearConstraint(
                    sp.lambdify(problem_args, c),
                    0,
                    np.inf,
                    jac=sp.lambdify(problem_args, c_jac),
                    hess="cs",
                )
            out_constriants.append(this_constraint)

        out_symbols = evaluate_inner(tape, arguments, outputs, self, workspace)

        out_map = [sp.lambdify(problem_args, o) for o in out_symbols]

        def solver(*solver_args):
            x0 = np.zeros(n)

            soln = sy.optimize.minimize(
                cost_f,
                x0,
                method="trust-constr",
                jac=cost_jac,
                hess=cost_hess,
                constraints=out_constriants,
            )

            inner_args = [soln.x] + list(*solver_args)
            result = [o(*inner_args)[0] for o in out_map]
            return [self.to_backend_array(r) for r in result]

        return solver


#
# OP v_1, v_2, v_3
#
# v is either
# a) a symbol
# b) a constant
# c) another node in the graph
#
# if OP is linear
# -



import numpy as np
from coker.algebra.kernel import TraceContext


def zeros(shape: tuple):
    tape = TraceContext.get_local_tape()
    assert tape is not None
    return tape.insert_value(np.zeros(shape))



from typing import Tuple, Type, List, Dict
from coker import Tracer, Function, OP
from coker.algebra.dimensions import FunctionSpace
from coker.backends.backend import Backend, ArrayLike, get_backend_by_name
from coker.backends.coker.ast_preprocessing import SparseNet
from coker.backends.coker.layers import (
    MemorySpec,
    InputLayer,
    OutputLayer,
    GenericLayerOP,
)
from coker.backends.coker.weights import BilinearWeights
from coker.backends.coker.op_impl import ops
import numpy as np
import scipy.sparse


class CokerBackend(Backend):
    def __init__(self):
        pass

    def to_backend_array(self, array: ArrayLike):
        pass

    def to_numpy_array(self, array) -> ArrayLike:
        pass

    def native_types(self) -> Tuple[Type]:
        pass

    def call(self, op, *args) -> ArrayLike:
        pass

    def evaluate(self, function, inputs: ArrayLike):

        if all(o is None for o in function.output):
            return function.output

        # Fall back to numpy for higher-order functions with FunctionSpace inputs.
        if any(
            isinstance(function.tape.dim[i], FunctionSpace)
            for i in function.tape.input_indicies
        ):
            numpy_backend = get_backend_by_name("numpy", set_current=False)
            return numpy_backend.evaluate(function, inputs)

        g = create_opgraph(function)

        return [g(*inputs)]

    def reshape(self, array: ArrayLike, shape: Tuple[int, ...]) -> ArrayLike:
        raise NotImplementedError

    def lower(self, function: Function):
        # FunctionSpace inputs and None outputs can't be lowered to a static
        # graph; fall back to the numpy plan for those cases.
        if any(
            isinstance(function.tape.dim[i], FunctionSpace)
            for i in function.tape.input_indicies
        ) or any(o is None for o in function.output):
            numpy_backend = get_backend_by_name("numpy", set_current=False)
            return numpy_backend.lower(function)

        g = create_opgraph(function)

        def compiled(inputs):
            return [g(*inputs)]

        return compiled

    def build_optimisation_problem(
        self,
        cost: Tracer,  # cost
        constraints: List[Tracer],
        parameters: List[Tracer],
        outputs: List[Tracer],
        initial_conditions: Dict[int, ArrayLike],
    ):
        raise NotImplementedError


def create_opgraph(function: Function):
    tape = function.tape

    input_layer = InputLayer()
    for i in tape.input_indicies:
        input_layer.add_input(tape.dim[i])

    input_memory = MemorySpec(location=0, count=input_layer.dimension)

    node_values = {
        i: BilinearWeights(
            linear=input_layer.get_projection(i),
            memory=input_memory,
            shape=tape.dim[i].shape,
        )
        for i in tape.input_indicies
    }

    layers = []
    next_location = input_layer.dimension
    output_indices = {
        output.index for output in function.output if output is not None
    }
    numpy_backend = get_backend_by_name("numpy", set_current=False)

    for idx in range(len(tape)):
        if idx in tape.input_indicies:
            continue

        op, *args = tape.nodes[idx]

        if op == OP.VALUE:
            (constant_value,) = args
            if scipy.sparse.issparse(constant_value):
                constant_value = constant_value.toarray()
            node_values[idx] = constant_value
            continue

        operands = [node_values[a.index] for a in args]
        # Collapse constant BilinearWeights to plain numpy arrays so the
        # constant-fold path below can consume them without needing a layer.
        operands = [
            (
                o.constant.toarray().reshape(o.shape)
                if isinstance(o, BilinearWeights) and o.is_constant
                else o
            )
            for o in operands
        ]
        bw_operands = [
            operand
            for operand in operands
            if isinstance(operand, BilinearWeights)
        ]

        if not bw_operands:
            node_values[idx] = numpy_backend.call(op, *operands)
            continue

        if idx not in output_indices and op in ops:
            memories = {id(operand.memory) for operand in bw_operands}
            if len(memories) == 1:
                try:
                    result = ops[op](*operands)
                    if isinstance(result, BilinearWeights):
                        node_values[idx] = result
                        continue
                except (AssertionError, TypeError, NotImplementedError):
                    pass

        count = tape.dim[idx].flat()
        mem = MemorySpec(location=next_location, count=count)
        next_location += count
        compiled_operands = [
            o.compile() if isinstance(o, BilinearWeights) else o
            for o in operands
        ]
        layers.append(GenericLayerOP(mem, op, *compiled_operands))
        shape = tape.dim[idx].shape or (1,)
        node_values[idx] = BilinearWeights.reshape_identity(
            memory=mem, shape=shape
        )

    output_layer = OutputLayer()
    for output in function.output:
        output_layer.add_output(node_values[output.index].memory, output.dim)

    return SparseNet([input_memory], input_layer, output_layer, layers)



import numpy as np
from coker import OP
from coker.backends.coker.sparse_tensor import (
    dok_ndarray,
    tensor_vector_product,
    is_constant,
)
from coker.backends.coker.tensor_contants import hat, levi_civita_3_tensor
from coker.backends.coker.weights import BilinearWeights


def cross(x, y):
    if is_constant(x):
        Ax = hat(x)
        return (Ax).toarray() @ y
    if is_constant(y):
        Ay = -hat(y)
        return Ay.toarray() @ x

    assert isinstance(x, BilinearWeights)
    assert isinstance(y, BilinearWeights)
    assert (
        x.memory == y.memory
    ), "cross product requires BilinearWeights from same memory"
    assert (
        x.is_linear and y.is_linear
    ), "cross product of quadratic weights not supported"

    eps = np.zeros((3, 3, 3))
    eps[0, 1, 2] = eps[1, 2, 0] = eps[2, 0, 1] = 1
    eps[2, 1, 0] = eps[0, 2, 1] = eps[1, 0, 2] = -1

    c_x = x.constant.toarray()  # shape (3,)
    c_y = y.constant.toarray()  # shape (3,)
    L_x = x.linear.toarray()  # shape (3, n)
    L_y = y.linear.toarray()  # shape (3, n)

    c_result = np.cross(c_x, c_y)  # shape (3,)
    # linear contribution: cross(L_x @ m, c_y) + cross(c_x, L_y @ m)
    #   = -hat(c_y) @ L_x @ m + hat(c_x) @ L_y @ m
    L_result = (
        -hat(c_y).toarray() @ L_x + hat(c_x).toarray() @ L_y
    )  # shape (3, n)
    # quadratic contribution: cross(L_x @ m, L_y @ m)
    Q_result = np.einsum("ijk,js,kt->ist", eps, L_x, L_y)  # shape (3, n, n)

    return BilinearWeights(
        x.memory,
        (3,),
        constant=dok_ndarray.fromarray(c_result),
        linear=dok_ndarray.fromarray(L_result),
        quadratic=dok_ndarray.fromarray(Q_result),
    )


def dot(x, y):
    if is_constant(x):
        if len(x.shape) == 1:
            (n,) = x.shape
            xt = x.T.reshape((1, n))
        else:
            n, m = x.shape
            assert m == 1
            xt = x.T
        return xt @ y
    if is_constant(y):
        if len(y.shape) == 1:
            (n,) = y.shape
            yT = y.T.reshape((1, n))
        else:
            n, m = y.shape
            assert m == 1
            yT = y.T
        return yT @ x

    if isinstance(x, BilinearWeights) and isinstance(y, BilinearWeights):
        return x.dot(y)

    raise NotImplementedError


def transpose(x):
    if is_constant(x):
        if len(x.shape) == 1:
            (n,) = x.shape
            return x.reshape((n, 1)).T

        return x.T

    if isinstance(x, BilinearWeights):
        return x.transpose()

    raise NotImplementedError(f"Cannot transpose {type(x)}, {x.shape}")


def is_scalar(x):
    if isinstance(x, (float, complex, int)):
        return True
    try:
        return all(s == 1 for s in x.shape)
    except AttributeError:
        pass
    if isinstance(x, BilinearWeights):
        return x.dimension == 1
    raise NotImplementedError


def mul(x, y):
    return x * y


def div(x, y):
    return x / y


ops = {
    OP.MUL: mul,
    OP.DIV: div,
    OP.ADD: lambda x, y: x + y,
    OP.SUB: lambda x, y: x - y,
    OP.NEG: lambda x: -x,
    OP.MATMUL: lambda x, y: x @ y,
    OP.CROSS: cross,
    OP.DOT: dot,
    OP.TRANSPOSE: transpose,
}



from coker import OP, Function
from coker.backends import get_backend_by_name


def try_rewrite_mul(nodes: list, atoms, constants, i):
    op, lhs, rhs = nodes[i]
    assert op == OP.MUL

    # case 1
    #  *(x,b )         ->  *(b, x)
    if lhs.index in atoms and rhs.index in constants:
        nodes[i] = (op, rhs, lhs)
        return

    # Case 2
    #  *(x,x)          ->  *(x, x)
    if lhs.index in atoms and rhs.index in atoms:
        return

    # Case 2b
    if lhs.index in constants and rhs.index in constants:
        constants.add(i)
        return

    # Case 3
    #  *((a, x), x)    -> *(a,  *(x, x))
    if lhs.index not in atoms and nodes[lhs.index][0] == op.MUL:
        _, parent_lhs, parent_rhs = nodes[lhs.index]
        #  n_lhs = *(a, x) -> n_lhs = a
        #
        if parent_lhs.index in constants and parent_rhs.index in atoms:
            nodes[lhs.index] = (op.MUL, parent_rhs, rhs)
            nodes[i] = (op, parent_lhs, lhs)

    # Case 4
    #  *(x, (b, x))    -> *(b,  *(x, x))
    elif rhs.index not in atoms and nodes[rhs.index][0] == op.MUL:
        _, parent_lhs, parent_rhs = nodes[rhs.index]
        if parent_lhs.index in constants and parent_rhs.index in atoms:
            nodes[rhs.index] = (op.MUL, lhs, parent_rhs)
            nodes[i] = (op, parent_lhs, rhs)

    # Case 5
    #  *((a, x), *(b,x)) -> *(ab, *(x,x))

    elif (
        rhs.index not in atoms
        and lhs.index in atoms
        and nodes[rhs.index][0] == op.MUL
        and nodes[lhs.index][0] == op.MUL
    ):
        _, pll, plr = nodes[lhs.index]
        _, prl, prr = nodes[rhs.index]
        if pll.index in constants and prl.index in constants:
            nodes[lhs.index] = (
                OP.VALUE,
                constants[pll.index] * constants[prl.index],
            )
            nodes[rhs.index] = (OP.MUL, plr, prr)

    #  *(?, x))     -> *(x, ?)      (polynomails)
    #  *(?, ?)      -> *(?,?)       (more polynomials)


def rewrite_graph(function: Function):
    constants = {}

    outputs = {o.index for o in function.output}
    inputs = set(function.tape.input_indicies)
    atoms = inputs.copy()
    work_set = [i for i in range(len(function.tape.nodes)) if i not in inputs]
    nodes = function.tape.nodes
    for i in work_set:
        op, *args = nodes[i]
        if op == OP.VALUE:
            (arg,) = args
            constants[i] = arg
            continue

        if all([a.index in constants for a in args]):
            constants[i] = get_backend_by_name(
                "numpy", set_current=False
            ).call(op, *[constants[a.index] for a in args])
            continue

        if op == OP.MUL:
            try_rewrite_mul(nodes, atoms, constants, i)

    function.tape.nodes = nodes
    return function



import operator
from functools import reduce

import numpy as np
from typing import Optional, Dict, Tuple, Union, NewType

MultiIndex = NewType("MultiIndex", Union[Tuple[int, ...], int])

scalar = (float, int, np.int32, np.int64, np.float32, np.float64)
NDArray = NewType("NDArray", Union[np.ndarray, "dok_ndarray"])


def is_constant(a):
    return (
        isinstance(a, scalar)
        or isinstance(a, np.ndarray)
        or isinstance(a, dok_ndarray)
    )


def cast_vector(arg, dim):
    if len(arg.shape) == 1:
        if arg.shape[0] == dim:
            return arg
    elif len(arg.shape) == 2:
        if arg.shape[0] == dim and arg.shape[1] == 1:
            return arg.reshape((dim, 1))

    raise TypeError(f"Cannot cast {arg.shape} vector to a {(dim,)} vector")


class dok_ndarray(np.lib.mixins.NDArrayOperatorsMixin):
    """Sparse Multilinear map

    Understood as a vector of linear tensor fields.
    Some notation:

    - `e^j` always refers to the basis vectors for the V_i vector space (i.e. a column vector)
    - `E_i` always refers to the basis covectors for the V_j vector space (i.e. a row vector)


    M = \alpha^i_{jkm} E_i*e^j*e^k*e^m...

    Operations:
        - addition, subtraction, and scalar multiplication as usual
        - 'matmul' is tensor contraction along the last + first dimensions
           that is, if N @ M means that if
                N = n_{ab...di}E_a e^b ...e^d e^i
           and
                M = m_{ijk...}E_ie^je^k...
           then
                NM = (nm)_{ab...djk}E_a e^b ...  e^d * e^je^k ...
           with
                the coefficients (nm) being summed appropriately.

        -
    """

    def __init__(self, shape, data: Optional[Dict[MultiIndex, float]] = None):
        self.keys = data if data is not None else {}
        self.shape = shape

    def is_empty(self):
        return self.keys == {}

    def is_scalar(self):
        return all(s == 1 for s in self.shape)

    def __setitem__(self, key, value):
        if isinstance(key, int):
            key = (key,)
        if value != 0:
            self.keys[key] = value
        else:
            if key in self.keys:
                del self.keys[key]

    def __getitem__(self, key):
        if isinstance(key, int):
            key = (key,)
        try:
            return self.keys[key]
        except KeyError:
            return 0

    def clone(self) -> "dok_ndarray":
        return dok_ndarray(self.shape, self.keys.copy())

    def toarray(self):
        m = np.zeros(shape=self.shape)
        for k, v in self.keys.items():
            m[k] = v
        return m

    def swap_indices(self, i, j):
        assert i != j and i < len(self.shape) and j < len(self.shape)
        if i > j:
            i, j = j, i

        data = {}
        for k, v in self.keys.items():
            k_i, k_j = k[i], k[j]
            k_prime = (*k[0:i], k_j, *k[i + 1 : j], k_i, *k[j + 1 :])
            data[k_prime] = v
        s_i, s_j = self.shape[i], self.shape[j]
        shape = (
            *self.shape[0:i],
            s_j,
            *self.shape[i + 1 : j],
            s_i,
            *self.shape[j + 1 :],
        )

        return dok_ndarray(shape=shape, data=data)

    @staticmethod
    def zeros(shape):
        return dok_ndarray(shape)

    @staticmethod
    def eye(n):
        shape = (n, n)
        keys = {(i, i): 1 for i in range(n)}

        return dok_ndarray(shape, keys)

    @staticmethod
    def fromarray(other: np.ndarray):
        shape = other.shape

        keys = {}
        with np.nditer(
            other, op_flags=["readonly"], flags=["multi_index", "reduce_ok"]
        ) as it:
            for item in it:
                if item != 0:
                    index = tuple(it.multi_index)
                    keys[index] = float(item)

        return dok_ndarray(shape, keys)

    @staticmethod
    def from_maybe(
        arg: Optional[Union[np.ndarray, "dok_ndarray"]],
        expected_shape: Optional[Tuple[int, ...]] = None,
    ) -> "dok_ndarray":

        if isinstance(arg, np.ndarray):
            result = dok_ndarray.fromarray(arg)
            return result
        if isinstance(arg, dok_ndarray):

            return arg.clone()
        if arg is None and expected_shape is not None:
            return dok_ndarray(expected_shape)

        raise TypeError(
            f"Don't know how to turn {arg} into an array of shape {expected_shape}"
        )

    @property
    def T(self):

        if len(self.shape) != 2:
            raise NotImplementedError(
                "Can't unambiguously transpose a higher dimensional array"
            )
        return dok_ndarray(
            tuple(reversed(self.shape)),
            {(k2, k1): v for (k1, k2), v in self.keys.items()},
        )

    def __float__(self):
        if self.is_scalar():
            if len(self.keys) == 1:
                (v,) = self.keys.values()
                return float(v)
            else:
                assert len(self.keys) == 0
                return 0.0

        raise TypeError(f"Cannot cast a {self.shape} array to a float")

    def __neg__(self):
        keys = {k: -v for k, v in self.keys.items()}
        return dok_ndarray(self.shape, keys)

    def __mul__(self, other):
        if isinstance(other, (dok_ndarray, np.ndarray)):
            assert other.shape in {(1,), (1, 1)}
            other = float(other)
        elif isinstance(other, scalar):
            other = float(other)
        else:
            raise TypeError(f"Cannot multiply by {other}")

        if other == 0:
            return dok_ndarray(self.shape, {})

        keys = {k: v * other for k, v in self.keys.items()}
        return dok_ndarray(self.shape, keys)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __radd__(self, other):
        return self.__add__(other)

    def __add__(self, other):
        if not other:
            return self
        if isinstance(other, scalar):
            if all(s == 1 for s in self.shape):
                k = tuple(0 for _ in self.shape)
                if self.keys:
                    data = {k: self.keys[k] + other}
                else:
                    data = {k: other}
                return dok_ndarray(self.shape, data)

        if isinstance(other, dok_ndarray):
            assert (
                other.shape == self.shape
            ), f"Cannot add tensors of shape {other.shape} to {self.shape}"
            keys = self.keys.copy()
            for k, v in other.keys.items():
                if k in keys:
                    keys[k] += v
                else:
                    keys[k] = v
            return dok_ndarray(self.shape, keys)

        raise NotImplementedError(
            f"Cannot add {other} to a tensor of shape {self.shape}"
        )

    def __sub__(self, other):
        if isinstance(other, dok_ndarray):
            keys = self.keys.copy()
            for k, v in other.keys.items():
                if k in keys:
                    keys[k] -= v
                else:
                    keys[k] = -v
        else:
            raise NotImplementedError()
        return dok_ndarray(self.shape, keys)

    def evaluate(self, *vectors: NDArray) -> NDArray:
        #        n_vector = len(vectors)

        vectors = [
            cast_vector(v, s)
            for v, s in zip(vectors, self.shape[-len(vectors) :])
        ]
        data = {}
        for k, v in self.keys.items():
            out_key = k[: -len(vectors)]
            k_prod = k[-len(vectors) :]
            a_k = v * reduce(
                operator.mul, (vec[j] for vec, j in zip(vectors, k_prod)), 1
            )
            if out_key not in data:
                data[out_key] = a_k
            else:
                data[out_key] += a_k

        shape = self.shape[: -len(vectors)]
        return dok_ndarray(shape, data)

    def __matmul__(self, other: Union[NDArray, Tuple[NDArray]]):
        """SymbolicVector contraction

        it T = a_{ijk} e_i e^j e^k

        """
        if isinstance(other, tuple):
            return self.evaluate(*other)

        elif isinstance(other, (dok_ndarray, np.ndarray)):
            if isinstance(other, np.ndarray) and len(other.shape) == 1:
                # Contract along the last axis with a 1D vector — use tensor_vector_product
                # to avoid adding a spurious trailing dimension.
                if self.shape[-1] != other.shape[0]:
                    raise TypeError(
                        f"Cannot multiply {self.shape} @ {other.shape}"
                    )
                if not self.keys:
                    return dok_ndarray(self.shape[:-1], {})
                return tensor_vector_product(
                    self, other, axis=len(self.shape) - 1
                )

            if self.shape[-1] != other.shape[0]:
                raise TypeError(
                    f"Cannot multiply {self.shape} @ {other.shape}"
                )
        else:
            return other.__rmatmul__(self)
        shape = (*self.shape[:-1], *other.shape[1:])

        if not self.keys:
            return dok_ndarray(shape, {})

        assert len(self.shape) > 1

        if isinstance(other, np.ndarray):
            return tensor_ndarray_product(self, other)
        else:
            return tensor_sum(self, other, l_index=len(self.shape) - 1)

    def __rmatmul__(self, other):
        try:
            assert (
                other.shape[-1] == self.shape[0]
            ), f"Can't multiple {other.shape} x {self.shape}"
        except AssertionError as ex:
            raise ex

        shape = (*other.shape[:-1], *self.shape[1:])
        data = {}
        with np.nditer(
            other, flags=["multi_index"], op_flags=["readonly"]
        ) as it:
            for v in it:
                *key_1, i_1 = it.multi_index
                for (i_2, *key_2), v_2 in self.keys.items():
                    if i_1 != i_2:
                        continue
                    key = (*key_1, *key_2)
                    if key in data:
                        data[key] += v_2 * float(v)
                    else:
                        data[key] = v_2 * float(v)
        return dok_ndarray(shape, data)

    def __eq__(self, other):
        if isinstance(other, dok_ndarray):
            return self.shape == other.shape and self.keys == other.keys
        if isinstance(other, np.ndarray):
            raise NotImplementedError

        if isinstance(other, scalar) and self.is_scalar():
            return float(self) == float(other)

        raise NotImplementedError

    def __array_ufunc__(self, ufunc, method, args, out=None):

        if ufunc == np.matmul and method == "__call__":
            return self.__rmatmul__(args).toarray()
        if ufunc == np.dot and method == "__call__":
            return args.T @ self

        if ufunc == np.multiply and method == "__call__":
            if args == 1.0:
                return self
            if self.is_scalar() or isinstance(args, scalar):
                return self.__mul__(args)

        raise NotImplementedError

    def __truediv__(self, other):
        if isinstance(other, dok_ndarray):
            other = float(other)
        assert isinstance(other, scalar)
        assert other != 0

        return dok_ndarray(
            self.shape, {k: v / other for k, v in self.keys.items()}
        )

    def reshape(self, shape):
        if len(shape) > len(self.shape) and all(
            s_i == s_n for s_i, s_n in zip(shape, (*self.shape, 1))
        ):
            self.keys = {(*k, 0): v for k, v in self.keys.items()}
            self.shape = (*shape[:-1], *shape[1:])
            return

        if len(shape) == len(self.shape) and all(
            s_i >= s_n for s_i, s_n in zip(shape, self.shape)
        ):
            self.shape = shape
            return
        raise NotImplementedError(
            f"Don't know how to reshape a {self.shape} tensor into a {shape} tensor"
        )


def tensor_vector_product(tensor: dok_ndarray, vector: np.ndarray, axis=1):
    assert isinstance(axis, int) and 0 <= axis < len(tensor.shape)

    shape = tuple(s for i, s in enumerate(tensor.shape) if i != axis)

    new_data = {}
    for k, v in tensor.keys.items():
        i = k[axis]
        entry = float(v * vector[i])
        new_key = tuple(k_i for idx, k_i in enumerate(k) if idx != axis)
        if new_key in new_data:
            new_data[new_key] += entry
        else:
            new_data[new_key] = entry

    return dok_ndarray(shape, new_data)


def tensor_ndarray_product(
    tensor: dok_ndarray, array: np.ndarray, l_axis=-1, r_axis=0
):

    assert isinstance(array, np.ndarray) and isinstance(tensor, dok_ndarray)
    assert abs(l_axis) < len(tensor.shape)
    assert abs(r_axis) < len(array.shape)
    assert array.shape[r_axis] == tensor.shape[l_axis]

    if l_axis == -1:
        l_axis = len(tensor.shape) - 1

    if r_axis == -1:
        r_axis = len(array.shape) - 1

    l_shape = tuple(s for i, s in enumerate(tensor.shape) if i != l_axis)
    r_shape = tuple(s for i, s in enumerate(array.shape) if i != r_axis)
    shape = tuple((*l_shape, *r_shape))
    data = {}
    with np.nditer(array, flags=["multi_index"], op_flags=["readonly"]) as it:
        for x in it:
            x_idx_front = it.multi_index[0:r_axis]
            x_idx_tail = it.multi_index[r_axis + 1 :]
            keys = (
                k for k in tensor.keys if k[l_axis] == it.multi_index[r_axis]
            )
            for k in keys:
                k_l = k[0:l_axis]
                k_r = k[l_axis + 1 :]

                key = tuple((*k_l, *k_r, *x_idx_front, *x_idx_tail))
                if key in data:
                    data[key] += float(x) * tensor.keys[k]
                else:
                    data[key] = float(x) * tensor.keys[k]

    return dok_ndarray(shape, data)


def tensor_sum(lhs: dok_ndarray, rhs, l_index=0, r_index=0):
    assert lhs.shape[l_index] == rhs.shape[r_index]

    new_shape = (
        *[s for i, s in enumerate(lhs.shape) if i != l_index],
        *[s for i, s in enumerate(rhs.shape) if i != r_index],
    )

    new_data = {}
    for k_l, v_l in lhs.keys.items():
        for k_r, v_r in rhs.keys.items():
            if k_l[l_index] == k_r[r_index]:
                new_key = (
                    *k_l[0:l_index],
                    *k_l[l_index + 1 :],
                    *k_r[0:r_index],
                    *k_r[r_index + 1 :],
                )

                v = v_l * v_r
                if new_key in new_data:
                    new_data[new_key] += v
                else:
                    new_data[new_key] = v

    return dok_ndarray(new_shape, new_data)



from coker.toolkits.system_modelling.modelling import (
    Block,
    BlockSpec,
    Signal,
    Vector,
    Parameter,
    Variable,
    ValueType,
)
from typing import Callable


class SignalGenerator(Block):
    spec = BlockSpec(
        inputs=[],
        outputs=[Signal("y")],
    )

    def __init__(self, name, parent, func: Callable[[float], float]):

        Block.__init__(self, name, SignalGenerator.spec)
        self.func = func

    def __call__(self, t, *args) -> float:
        return self.func(t)

    def free_parameters(self) -> set:
        return set()


class TransferFunction(Block):
    def __init__(self, name, numerator: Vector, denominator: Vector):
        spec = BlockSpec(
            inputs=[Signal("u")],
            outputs=[Signal("y")],
            parameters=[Vector("numerator"), Vector("denominator")],
            state=[],
        )
        Block.__init__(self, name, spec)
        self.numerator = numerator
        self.denominator = denominator


class Step(Block):
    spec = BlockSpec(
        outputs=[Signal("y")],
        parameters=[Parameter("t_step")],
    )

    def __init__(self, name, parent, t_step=None):
        if t_step is None:
            super().__init__(name, parent, self.spec)
        else:
            super().__init__(
                name,
                parent,
                BlockSpec(
                    outputs=[Signal("y")],
                    parameters=[Parameter("t_step", t_step)],
                ),
            )
            self.t_step = t_step


class Integrator(Block):
    SPEC = BlockSpec(
        inputs=[Signal("u")],
        state=[Variable("x")],
        outputs=[Signal("y")],
    )

    def __init__(self, name=None):
        if name is None:
            name = "1/s"
        super().__init__(name, Integrator.SPEC)


class Gain(Block):
    def __init__(self, gain: float, name=None):
        if name is None:
            param = Parameter(name="gain", value=gain)
        else:
            param = Parameter(name=name, value=gain)
        spec = BlockSpec(
            inputs=[Signal("u")], outputs=[Signal("y")], parameters=[param]
        )

        super().__init__(name, spec)


class Constant(Block):
    def __init__(self, value: ValueType, name=None):
        spec = BlockSpec(
            inputs=[Signal("u")],
            outputs=[Signal("y")],
            parameters=[Parameter("value", value)],
        )

        super().__init__(name, spec)


class Difference(Block):
    SPEC = BlockSpec(
        inputs=[Signal("+"), Signal("-")],
        outputs=[Signal("y")],
    )

    def __init__(self, name=None):
        super().__init__(name, Difference.SPEC)


class Sum(Block):
    def __init__(self, name=None, inputs: int = 2):
        spec = BlockSpec(
            inputs=[Signal(f"u_{i}") for i in range(inputs)],
            outputs=[Signal("y")],
        )
        super().__init__(name, spec)



from typing import List, NamedTuple, Callable
import numpy as np

import coker
from coker.backends.backend import ArrayLike, Backend
from coker.algebra.kernel import Tracer, OP


# ---------------------------------------------------------------------------
# Compiled execution plan
# ---------------------------------------------------------------------------


class _PlanStep(NamedTuple):
    fn: Callable
    arg_indices: list
    out_idx: int
    dim: object
    post_fn: Callable  # pre-resolved reshape (or identity) applied after fn


class CompiledPlan:
    """Pre-compiled execution plan for a (tape, backend) pair.

    Built once via _build_plan; subsequent calls skip per-node isinstance
    dispatch and dict lookups by working from pre-resolved callables and
    workspace indices. Not thread-safe — workspace is mutated in place.
    """

    def __init__(self, steps, workspace, input_indices):
        self._steps = steps
        self._workspace = workspace  # constants pre-filled; reused each call
        self._input_indices = input_indices

    def execute(self, inputs, backend):
        ws = self._workspace
        ws[-1] = None

        for ws_idx, arg in zip(self._input_indices, inputs):
            if ws_idx >= 0:
                ws[ws_idx] = (
                    arg
                    if isinstance(arg, coker.Function)
                    else backend.to_backend_array(arg)
                )

        for step in self._steps:
            ws[step.out_idx] = step.post_fn(
                step.fn(*[ws[i] for i in step.arg_indices])
            )

        return ws


def _build_plan(graph, backend):
    """Walk the tape once and return a CompiledPlan."""

    # Pass 1 — mark which nodes depend on inputs (dynamic) vs are pure constants.
    is_dynamic = {}
    for i, node in enumerate(graph.nodes):
        if isinstance(node, Tracer):
            # Bare Tracer in nodes list means this is an input node.
            is_dynamic[i] = True
        else:
            op, *args = node
            is_dynamic[i] = any(
                isinstance(a, Tracer)
                and a.tape is graph
                and is_dynamic.get(a.index, False)
                for a in args
            )

    # Pass 2 — pre-evaluate constant nodes into the workspace.
    # Negative slots below -(len+10) are reserved for any inline constants
    # (cross-tape Tracers or bare values) that appear as node arguments.
    workspace = {-1: None}
    next_slot = [-(len(graph.nodes) + 10)]

    def alloc_inline(value):
        s = next_slot[0]
        next_slot[0] -= 1
        workspace[s] = value
        return s

    for i, node in enumerate(graph.nodes):
        if is_dynamic.get(i, True) or isinstance(node, Tracer):
            continue
        op, *args = node
        resolved = []
        for a in args:
            if isinstance(a, Tracer) and a.tape is graph:
                resolved.append(workspace[a.index])
            elif isinstance(a, Tracer):
                resolved.append(a)  # cross-tape: pass through as-is
            else:
                resolved.append(backend.to_backend_array(a))
        value = resolved[0] if op == OP.VALUE else backend.call(op, *resolved)
        if not isinstance(value, Tracer):
            value = backend.reshape(value, graph.dim[i])
        workspace[i] = value

    # Pass 3 — build execution steps for dynamic non-input nodes only.
    steps = []
    for i, node in enumerate(graph.nodes):
        if not is_dynamic.get(i, False) or isinstance(node, Tracer):
            continue
        op, *args = node
        arg_indices = []
        for a in args:
            if isinstance(a, Tracer) and a.tape is graph:
                arg_indices.append(a.index)
            elif isinstance(a, Tracer):
                arg_indices.append(alloc_inline(a))
            else:
                arg_indices.append(alloc_inline(backend.to_backend_array(a)))
        dim = graph.dim[i]
        steps.append(
            _PlanStep(
                backend.resolve_fn(op),
                arg_indices,
                i,
                dim,
                backend.resolve_post_fn(dim),
            )
        )

    return CompiledPlan(steps, workspace, graph.input_indicies)


def _cast_outputs(outputs, graph, workspace, backend):
    """Extract and reshape outputs from the workspace after plan execution."""
    result = []
    for o in outputs:
        if o is None:
            result.append(None)
            continue
        if o.tape is not graph:
            result.append(o)
            continue
        output = workspace[o.index]
        if isinstance(output, Tracer):
            result.append(output)
            continue
        if not o.dim.is_scalar():
            try:
                output = backend.to_numpy_array(output)
                if isinstance(output, np.ndarray):
                    result.append(np.reshape(output, o.shape))
                    continue
            except ValueError:
                pass
            backend.reshape(output, o.dim)
            result.append(output)
            continue
        result.append(backend.to_numpy_array(output))
    return result


# ---------------------------------------------------------------------------
# Original interpreted evaluator (kept for sympy backend and optimisation)
# ---------------------------------------------------------------------------


def evaluate_inner(graph, args, outputs, backend: Backend, workspace: dict):
    workspace[-1] = None
    for index, arg in zip(graph.input_indicies, args):
        if isinstance(arg, coker.Function):
            workspace[index] = arg
        else:
            workspace[index] = backend.to_backend_array(arg)

    work_list = [i for i in range(len(graph.nodes)) if i not in workspace]

    def cast_node(node):
        try:
            if isinstance(node, Tracer):
                if node.tape == graph:
                    return workspace[node.index]
                else:
                    return node
            elif isinstance(node, coker.Function):
                return node

            return backend.to_backend_array(node)
        except Exception as ex:
            ex.add_note(f"Node index: {node.index}")
            raise ex from ex

    for w in work_list:
        op, *nodes = graph.nodes[w]

        args = [cast_node(n) for n in nodes]
        if op == OP.VALUE:
            (value,) = args
        else:
            try:
                value = backend.call(op, *args)
            except Exception as ex:
                ex.add_note(f"Node index: {w}")
                ex.add_note(f"Node: {op}({args})")
                raise ex from ex

        workspace[w] = (
            backend.reshape(value, graph.dim[w])
            if not isinstance(value, Tracer)
            else value
        )

    def cast_output(o):
        if o is None:
            return None
        if o.tape != graph:
            return o
        output = workspace[o.index]
        if isinstance(output, Tracer):
            return output
        if not o.dim.is_scalar():
            try:
                output = backend.to_numpy_array(output)
                if isinstance(output, np.ndarray):
                    return np.reshape(output, shape=o.shape)
            except ValueError:
                pass
            backend.reshape(output, o.dim)
            return output
        return backend.to_numpy_array(workspace[o.index])

    outputs = [cast_output(o) for o in outputs]

    return outputs


def evaluate(function, args, backend=None):

    from coker.backends import get_backend_by_name, get_current_backend

    if not backend:
        backend_impl: Backend = get_current_backend()
    else:
        backend_impl: Backend = get_backend_by_name(backend)
    return backend_impl.evaluate(function, args)



from matplotlib import pyplot as plt
import numpy as np
from typing import List
from coker.toolkits.kinematics.rigid_body import RigidBody
from coker.toolkits.spatial import Isometry3

from mpl_toolkits.mplot3d.art3d import Line3D
from matplotlib.animation import FuncAnimation

e_x = np.array([1, 0, 0])
e_y = np.array([0, 1, 0])
e_z = np.array([0, 0, 1])


class JointWidget:
    def __init__(self, transform: Isometry3, scale):
        self.x_axis = Line3D([], [], [], color="blue")
        self.y_axis = Line3D([], [], [], color="green")
        self.z_axis = Line3D([], [], [], color="red")
        self.point = Line3D(
            [], [], [], marker="o", markersize=1, markerfacecolor="black"
        )
        self.scale = scale
        self.set_transform(transform)

    def set_transform(self, transform: Isometry3):
        point = transform.apply(
            np.zeros(
                3,
            )
        )

        x, y, z = point.tolist()
        self.point.set_data_3d([x], [y], [z])

        dx = transform.apply(e_x) - point
        dy = transform.apply(e_y) - point
        dz = transform.apply(e_z) - point
        x_x, x_y, x_z = (point + self.scale * dx).tolist()
        y_x, y_y, y_z = (point + self.scale * dy).tolist()
        z_x, z_y, z_z = (point + self.scale * dz).tolist()

        self.x_axis.set_data_3d([x, x_x], [y, x_y], [z, x_z])
        self.y_axis.set_data_3d([x, y_x], [y, y_y], [z, y_z])
        self.z_axis.set_data_3d([x, z_x], [y, z_y], [z, z_z])

        return self.point, self.x_axis, self.y_axis, self.z_axis

    def get_artists(self):
        return self.point, self.x_axis, self.y_axis, self.z_axis


class CenterOfMassWidget:
    def __init__(self, parent: int, transform: Isometry3):
        self.parent = parent
        self.transform = transform
        self.artist = Line3D([], [], [], linestyle="", marker="o")

    def set_transform(self, joint_transforms: List[Isometry3]):
        transform = joint_transforms[self.parent] @ self.transform

        x, y, z = transform.apply(np.zeros(3)).tolist()
        self.artist.set_data_3d([x], [y], [z])

    def get_artists(self):
        return self.artist


class LimbWidget:
    def __init__(self, path: List[int], effector_xform: Isometry3):
        self.path = path
        self.line = Line3D([], [], [], color="black")
        self.xform = effector_xform
        self.bounds = []

    def get_artists(self):
        return self.line

    def set_transform(self, transforms: List[Isometry3]):
        xforms = [transforms[i] for i in self.path]
        xforms.append(xforms[-1] @ self.xform)

        x, y, z = zip(*(t.apply(np.zeros((3,))).tolist() for t in xforms))

        self.line.set_data_3d(x, y, z)

        return self.line


class KinematicsVisualiser:
    def __init__(self, model: RigidBody, scale=0.25):
        self.model = model
        self.figure = plt.figure()
        ax = self.figure.add_subplot(projection="3d", aspect="equal")

        rest_config = np.zeros(model.total_joints())
        self.masses = [
            CenterOfMassWidget(i, inertia.centre_of_mass)
            for i, inertia in enumerate(model.inertia)
        ]

        self.joints = [
            JointWidget(t, scale) for t in model.joint_transforms(rest_config)
        ]

        self.end_effector_coords = [
            JointWidget(t, scale)
            for t in model.forward_kinematics(rest_config)
        ]

        for mass in self.masses:
            ax.add_artist(mass.get_artists())

        for joints in self.joints:
            [ax.add_artist(a) for a in joints.get_artists()]

        self.end_effectors = [
            LimbWidget([*model.get_dependent_links(parent)], xform)
            for i, (parent, xform) in enumerate(model.end_effectors)
        ]

        for end_effectors in self.end_effector_coords:
            [ax.add_artist(a) for a in end_effectors.get_artists()]

        locations = model.joint_transforms(rest_config)
        for end_effectors in self.end_effectors:
            end_effectors.set_transform(locations)
            ax.add_artist(end_effectors.get_artists())

        self.ax = ax
        self.update_values(rest_config)
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        ax.set_zlabel("z")

    def update_values(self, angles):
        joint_locations = self.model.joint_transforms(angles)
        fk = self.model.forward_kinematics(angles)

        for joint, transform in zip(self.joints, joint_locations):
            joint.set_transform(transform)

        for effector, transform in zip(self.end_effector_coords, fk):
            effector.set_transform(transform)

        for end_effector in self.end_effectors:
            end_effector.set_transform(joint_locations)

        for mass in self.masses:
            mass.set_transform(joint_locations)

    def get_artists(self):
        return (
            [j.get_artists() for j in self.joints]
            + [e.get_artists() for e in self.end_effectors]
            + [e.get_artists() for e in self.end_effector_coords]
            + [m.get_artists() for m in self.masses]
        )

    def set_view(self, x_lim=None, y_lim=None, z_lim=None):
        if x_lim is not None:
            self.ax.set_xlim(*x_lim)
        if y_lim is not None:
            self.ax.set_ylim(*y_lim)
        if z_lim is not None:
            self.ax.set_zlim(*z_lim)
        self.ax.set_aspect("equal")

    def animate_sweep(self, angles_over_time, loop=False, interval=100):
        def update(frame):
            angles = angles_over_time[frame]
            self.update_values(angles)
            return self.get_artists()

        ani = FuncAnimation(
            fig=self.figure,
            func=update,
            frames=len(angles_over_time),
            interval=interval,
            repeat=loop,
        )
        plt.show()
